#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Defines classes and methods used by example_parallel_network.py script

Copyright (C) 2012 Computational Neuroscience Group, NMBU.

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

"""

import numpy as np
import os
import scipy.stats as stats
import h5py
from mpi4py import MPI
import neuron
from .templatecell import TemplateCell
import scipy.sparse as ss

# set up MPI environment
COMM = MPI.COMM_WORLD
SIZE = COMM.Get_size()
RANK = COMM.Get_rank()


flattenlist = lambda lst: [item for sublist in lst for item in sublist]


################################################################################
# NetworkCell class that has a create_synapse method that
# creates a synapse on the target cell, and a create_spike_detector method that
# allows for connecting to a synapse on a target cell. All other methods and
# attributes are inherited from the standard LFPy.TemplateCell class
################################################################################
class NetworkCell(TemplateCell):
    """
    class NetworkCell

    Similar to `LFPy.TemplateCell` with the addition of some attributes and
    methods allowing for spike communication between parallel RANKs.

    This class allow using NEURON templates with some limitations.

    This takes all the same parameters as the Cell class, but requires three
    more template related parameters

    Parameters
    ----------
    morphology : str
        path to morphology file
    templatefile : str
        File with cell template definition(s)
    templatename : str
        Cell template-name used for this cell object
    templateargs : str
        Parameters provided to template-definition
    v_init : float
        Initial membrane potential. Default to -65.
    Ra : float
        axial resistance. Defaults to 150.
    cm : float
        membrane capacitance. Defaults to 1.0
    passive : bool
        Passive mechanisms are initialized if True. Defaults to True
    passive_parameters : dict
        parameter dictionary with values for the passive membrane mechanism in
        NEURON ('pas'). The dictionary must contain keys 'g_pas' and 'e_pas',
        like the default: passive_parameters=dict(g_pas=0.001, e_pas=-70)
    extracellular : bool
        switch for NEURON's extracellular mechanism. Defaults to False
    dt: float
        Simulation time step. Defaults to 2**-4
    tstart : float
        initialization time for simulation <= 0 ms. Defaults to 0.
    tstop : float
        stop time for simulation > 0 ms. Defaults to 100.
    nsegs_method : 'lambda100' or 'lambda_f' or 'fixed_length' or None
        nseg rule, used by NEURON to determine number of compartments.
        Defaults to 'lambda100'
    max_nsegs_length : float or None
        max segment length for method 'fixed_length'. Defaults to None
    lambda_f : int
        AC frequency for method 'lambda_f'. Defaults to 100
    d_lambda : float
        parameter for d_lambda rule. Defaults to 0.1
    delete_sections : bool
        delete pre-existing section-references. Defaults to True
    custom_code : list or None
        list of model-specific code files ([.py/.hoc]). Defaults to None
    custom_fun : list or None
        list of model-specific functions with args. Defaults to None
    custom_fun_args : list or None
        list of args passed to custom_fun functions. Defaults to None
    pt3d : bool
        use pt3d-info of the cell geometries switch. Defaults to False
    celsius : float or None
        Temperature in celsius. If nothing is specified here
        or in custom code it is 6.3 celcius
    verbose : bool
        verbose output switch. Defaults to False

    Examples
    --------

    >>> import LFPy
    >>> cellParameters = {
    >>>     'morphology' : '<path to morphology.hoc>',
    >>>     'templatefile' :  '<path to template_file.hoc>',
    >>>     'templatename' :  'templatename',
    >>>     'templateargs' :  None,
    >>>     'v_init' : -65,
    >>>     'cm' : 1.0,
    >>>     'Ra' : 150,
    >>>     'passive' : True,
    >>>     'passive_parameters' : {'g_pas' : 0.001, 'e_pas' : -65.},
    >>>     'dt' : 2**-3,
    >>>     'tstart' : 0,
    >>>     'tstop' : 50,
    >>> }
    >>> cell = LFPy.NetworkCell(**cellParameters)
    >>> cell.simulate()


    """
    def __init__(self, **args):
        """
        Initialization of class LFPy.NetworkCell.

        """
        TemplateCell.__init__(self, **args)

        # create list netconlist for spike detecting NetCon object(s)
        self.sd_netconlist = neuron.h.List()
        # create list of recording device for action potentials
        self.spikes = []
        # create list of random number generators used with synapse model
        self.rng_list = []

        # create separate list for networked synapses
        self.netconsynapses = []

        # create recording device for membrane voltage
        self.somav = neuron.h.Vector()
        for sec in self.somalist:
            self.somav.record(sec(0.5)._ref_v)


    def create_synapse(self, cell, sec, x=0.5, syntype=neuron.h.ExpSyn,
                       synparams=dict(tau=2., e=0.),
                       assert_syn_values=False):
        """
        Create synapse object of type syntype on sec(x) of cell and
        append to list cell.netconsynapses

        TODO: Use LFPy.Synapse class if possible.

        Parameters
        ----------
        cell : object
            instantiation of class NetworkCell or similar
        sec : neuron.h.Section object,
            section reference on cell
        x : float in [0, 1],
            relative position along section
        syntype : hoc.HocObject
            NEURON synapse model reference, e.g., neuron.h.ExpSyn
        synparams : dict
            parameters for syntype, e.g., for neuron.h.ExpSyn we have:
                tau : float, synapse time constant
                e : float, synapse reversal potential
        assert_syn_values : bool
            if True, raise AssertionError if synapse attribute values do not
            match the values in the synparams dictionary

        Raises
        ------
        AssertionError
        """
        # create a synapse object on the target cell
        syn = syntype(x, sec=sec)
        if hasattr(syn, 'setRNG'):
            # Create the random number generator for the synapse
            rng = neuron.h.Random()
            # not sure if this is how it is supposed to be set up...
            rng.MCellRan4(np.random.randint(0, 2**32-1), np.random.randint(0, 2**32-1))
            rng.uniform(0, 1)
            syn.setRNG(rng) # used for e.g., stochastic synapse mechanisms (cf. BBP microcircuit portal files)
            cell.rng_list.append(rng) # must store ref to rng object
        cell.netconsynapses.append(syntype(x, sec=sec))

        for key, value in synparams.items():
            exec("cell.netconsynapses[-1].{} = {}".format(key, value))
            # check that synapses are parameterized correctly
            if assert_syn_values:
                try:
                    np.testing.assert_almost_equal(getattr(cell.netconsynapses[-1], key), value)
                except AssertionError:
                    raise AssertionError('{} = {} != {}'.format(key,
                                                                getattr(cell.netconsynapses[-1], key),
                                                                value))


    def create_spike_detector(self, target=None, threshold=-10.,
                     weight=0.0, delay=0.0):
        """
        Create spike-detecting NetCon object attached to the cell's soma
        midpoint, but this could be extended to having multiple spike-detection
        sites. The NetCon object created is attached to the cell's sd_netconlist
        attribute, and will be used by the Network class when creating
        connections between all presynaptic cells and postsynaptic cells on
        each local RANK.

        Parameters
        ----------
        target : None (default) or a NEURON point process
        threshold : float
            spike detection threshold
        weight : float
            connection weight (not used unless target is a point process)
        delay : float
            connection delay (not used unless target is a point process)
        """
        # create new NetCon objects for the connections. Activation times will
        # be triggered on the somatic voltage with a given threshold.
        for sec in self.somalist:
            self.sd_netconlist.append(neuron.h.NetCon(sec(0.5)._ref_v,
                                                target,
                                                sec=sec))
            self.sd_netconlist[-1].threshold = threshold
            self.sd_netconlist[-1].weight[0] = weight
            self.sd_netconlist[-1].delay = delay


class DummyCell(object):
    def __init__(self, totnsegs=0,
                 imem=np.array([[]]),
                 xstart=np.array([]), xmid=np.array([]), xend=np.array([]),
                 ystart=np.array([]), ymid=np.array([]), yend=np.array([]),
                 zstart=np.array([]), zmid=np.array([]), zend=np.array([]),
                 diam=np.array([]), area=np.array([]), somainds=np.array([])):
        """
        Dummy Cell object initialized with all attributes needed for LFP
        calculations using the LFPy.RecExtElectrode class and methods. This cell
        can be imagined as one "super" cell containing transmembrane currents
        generated by all NetworkCell segments on this RANK at once.


        Parameters
        ----------
        totnsegs : int
            total number of segments
        imem : ndarray
            totnsegs x ntimesteps array with transmembrane currents in nA
        xstart, ystart, zstart : ndarray
            arrays of length totnsegs with start (x,y,z) coordinate of segments
            in units of um
        xmid, ymid, zmid : ndarray
            midpoint coordinates of segments
        xend, yend, zend : ndarray
            endpoint coordinateso of segments
        diam : ndarray
            array of length totnsegs with segment diameters
        area : ndarray
            array of segment surface areas
        """
        # set attributes
        self.totnsegs = totnsegs
        self.imem = imem
        self.xstart = xstart
        self.xmid = xmid
        self.xend = xend
        self.ystart = ystart
        self.ymid = ymid
        self.yend = yend
        self.zstart = zstart
        self.zmid = zmid
        self.zend = zend
        self.diam = diam
        self.area = area
        self.somainds = somainds

    def get_idx(self, section="soma"):
        if section=="soma":
            return self.somainds
        else:
            raise ValueError('section argument must be "soma"')


class NetworkPopulation(object):
    def __init__(self, CWD=None, CELLPATH=None, first_gid=0, Cell=NetworkCell,
                 POP_SIZE=4, name='L5PC',
                 cell_args=dict(), pop_args=dict(),
                 rotation_args=dict(),
                 OUTPUTPATH='example_parallel_network'):
        """
        NetworkPopulation class representing a group of Cell objects distributed
        across RANKs.

        Parameters
        ----------
        CWD : path or None
            Current working directory
        CELLPATH: path or None
            Relative path from CWD to source files for cell model (morphology, hoc routines etc.)
        first_gid : int
            The global identifier of the first cell created in this population
            instance. The first_gid in the first population created should be 0
            and cannot exist in previously created NetworkPopulation instances
        Cell : class
            class defining a Cell object, see class NetworkCell above
        POP_SIZE : int
            number of cells in population
        name : str
            population name reference
        cell_args : dict
            keys and values for Cell object
        pop_args : dict
            keys and values for Network.draw_rand_pos assigning cell positions
        rotation_arg : dict
            default cell rotations around x and y axis on the form
            { 'x' : np.pi/2, 'y' : 0 }. Can only have the keys 'x' and 'y'.
            Cells are randomly rotated around z-axis using the Cell.set_rotation
            method.
        OUTPUTPATH : str
            path to output file destination
        """
        # set class attributes
        self.CWD = CWD
        self.CELLPATH = CELLPATH
        self.first_gid = first_gid
        self.Cell = Cell
        self.POP_SIZE = POP_SIZE
        self.name = name
        self.cell_args = cell_args
        self.pop_args = pop_args
        self.rotation_args = rotation_args
        self.OUTPUTPATH = OUTPUTPATH

        # create folder for output if it does not exist
        if RANK == 0:
            if not os.path.isdir(OUTPUTPATH):
                os.mkdir(OUTPUTPATH)
        COMM.Barrier()

        # container of Vector objects used to record times of action potentials
        self.spike_vectors = []

        # set up population of cells on this RANK
        self.gids = [(i+first_gid) for i in range(POP_SIZE) if (i+first_gid) % SIZE == RANK]

        # we have to enter the cell's corresponding file directory to
        # create cell because how EPFL set their code up
        if CWD is not None:
            os.chdir(os.path.join(CWD, CELLPATH, self.name))
            self.cells = [Cell(**cell_args) for gid in self.gids]
            os.chdir(CWD)
        else:
            self.cells = [Cell(**cell_args) for gid in self.gids]
        # position each cell's soma in space
        self.soma_pos = self.draw_rand_pos(POP_SIZE=len(self.gids), **pop_args)
        for i, cell in enumerate(self.cells):
            cell.set_pos(**self.soma_pos[i])

        # assign a random rotation around the z-axis of each cell
        self.rotations = np.random.uniform(0, np.pi*2, len(self.gids))
        assert('z' not in self.rotation_args.keys())
        for i, cell in enumerate(self.cells):
            cell.set_rotation(z=self.rotations[i], **self.rotation_args)

        # assign gid to each cell
        for gid, cell in zip(self.gids, self.cells):
            cell.gid = gid


        # gather gids, soma positions and cell rotations to RANK 0, and write
        # as structured array.
        if RANK == 0:
            populationData = flattenlist(COMM.gather(zip(self.gids, self.soma_pos, self.rotations)))

            # create structured array for storing data
            dtype = [('gid', 'i8'), ('x', float), ('y', float), ('z', float),
                     ('x_rot', float), ('y_rot', float), ('z_rot', float)]
            popDataArray = np.empty((len(populationData, )), dtype=dtype)
            for i, (gid, pos, z_rot) in enumerate(populationData):
                popDataArray[i]['gid'] = gid
                popDataArray[i]['x'] = pos['x']
                popDataArray[i]['y'] = pos['y']
                popDataArray[i]['z'] = pos['z']
                popDataArray[i]['x_rot'] = np.pi/2
                popDataArray[i]['y_rot'] = 0.
                popDataArray[i]['z_rot'] = z_rot

            # Dump to hdf5 file, append to file if it exists
            f = h5py.File(os.path.join(self.OUTPUTPATH,
                                       'cell_positions_and_rotations.h5'), 'a')
            # delete old entry if it exist
            if self.name in f.keys():
                del f[self.name]
                try:
                    assert self.name not in f.keys()
                except AssertionError:
                    raise AssertionError
            f[self.name] = popDataArray
            f.close()
        else:
            COMM.gather(zip(self.gids, self.soma_pos, self.rotations))

        # sync
        COMM.Barrier()


    def draw_rand_pos(self, POP_SIZE, radius, loc, scale, cap=None):
        """
        Draw some random location for POP_SIZE cells within radius radius,
        at mean depth loc and standard deviation scale.

        Returned argument is a list of dicts [{'x', 'y', 'z'},].


        Parameters
        ----------
        POP_SIZE : int
            Population size
        radius : float
            Radius of population.
        loc : float
            expected mean depth of somas of population.
        scale : float
            expected standard deviation of depth of somas of population.
        cap : None, float or length to list of floats
            if float, cap distribution between [loc-cap, loc+cap),
            if list, cap distribution between [loc-cap[0], loc+cap[1]]


        Returns
        -------
        soma_pos : list
            List of dicts of len POP_SIZE
            where dict have keys x, y, z specifying
            xyz-coordinates of cell at list entry `i`.


        """

        x = np.empty(POP_SIZE)
        y = np.empty(POP_SIZE)
        z = np.empty(POP_SIZE)
        for i in range(POP_SIZE):
            x[i] = (np.random.rand()-0.5) * radius*2
            y[i] = (np.random.rand()-0.5) * radius*2
            while np.sqrt(x[i]**2 + y[i]**2) >= radius:
                x[i] = (np.random.rand()-0.5)*radius*2
                y[i] = (np.random.rand()-0.5)*radius*2
        z = np.random.normal(loc=loc, scale=scale, size=POP_SIZE)
        if cap is not None:
            if type(cap) in [float, np.float, np.float32, np.float64]:
                while not np.all((z >= loc-cap) & (z < loc+cap)):
                    inds = (z < loc-cap) ^ (z > loc+cap)
                    z[inds] = np.random.normal(loc=loc, scale=scale,
                                               size=inds.sum())
            elif type(cap) is list:
                try:
                    assert(len(cap) == 2)
                except AssertionError:
                    raise AssertionError('cap = {} is not a length 2 list'.format(float))
                while not np.all((z >= loc-cap[0]) & (z < loc+cap[1])):
                    inds = (z < loc-cap[0]) ^ (z > loc+cap[1])
                    z[inds] = np.random.normal(loc=loc, scale=scale,
                                               size=inds.sum())
            else:
                raise Exception('cap = {} is not None, a float or length 2 list of floats'.format(float))

        soma_pos = []
        for i in range(POP_SIZE):
            soma_pos.append({'x' : x[i], 'y' : y[i], 'z' : z[i]})

        return soma_pos


class Network(object):
    def __init__(self, dt=0.1, tstart=0., tstop=1000., v_init=-65., celsius=6.3,
                 OUTPUTPATH='example_parallel_network',
                 verbose=False):
        """
        Network class, creating distributed populations of cells of
        type Cell and handling connections between cells in the respective
        populations.

        Parameters
        ----------
        dt : float
            Simulation timestep size
        tstart : float
            Start time of simulation
        tstop : float
            End time of simulation
        v_init : float
            Membrane potential set at first timestep across all cells
        celsius : float
            Global control of temperature, affect channel kinetics.
            It will also be forced when creating the different Cell objects, as
            LFPy.Cell and LFPy.TemplateCell also accept the same keyword
            argument.
        verbose : bool
            if True, print out misc. messages


        """
        # set attributes
        self.dt = dt
        self.tstart = tstart
        self.tstop = tstop
        self.v_init = v_init
        self.celsius = celsius
        self.OUTPUTPATH = OUTPUTPATH
        self.verbose = verbose

        # we need NEURON's ParallelContext for communicating NetCon events
        self.pc = neuron.h.ParallelContext()

        # create empty list for connections between cells (not to be confused
        # with each cell's list of netcons)
        self.netconlist = neuron.h.List()


        # The different populations in the Network will be collected in
        # a dictionary of NetworkPopulation object, where the keys represent the
        # population name. The names are also put in a list ordered according to
        # order populations are created in (as some operations rely on this)
        self.populations = dict()
        self.population_names = []



    def create_population(self, CWD=None, CELLPATH=None, Cell=NetworkCell,
                          POP_SIZE=4, name='L5PC',
                          cell_args=dict(), pop_args=dict(),
                          rotation_args=dict()):
        """
        Create and append a distributed POP_SIZE-sized population of cells of
        type Cell with the corresponding name. Cell-object references, gids on
        this RANK, population size POP_SIZE and names will be added to the lists
        Network.gids, Network.cells, Network.sizes and Network.names,
        respectively

        Parameters
        ----------
        CWD : path
            Current working directory
        CELLPATH: path
            Relative path from CWD to source files for cell model (morphology, hoc routines etc.)
        Cell : class
            class defining a Cell-like object, see class NetworkCell
        POP_SIZE : int
            number of cells in population
        name : str
            population name reference
        cell_args : dict
            keys and values for Cell object
        pop_args : dict
            keys and values for Network.draw_rand_pos assigning cell positions
        rotation_arg : dict
            default cell rotations around x and y axis on the form
            { 'x' : np.pi/2, 'y' : 0 }. Can only have the keys 'x' and 'y'.
            Cells are randomly rotated around z-axis using the Cell.set_rotation
            method.

        """
        try:
            assert name not in self.populations.keys()
        except AssertionError:
            raise AssertionError('population name {} already taken'.format(name))

        # compute the first global id of this new population, based
        # on population sizes of existing populations
        first_gid = 0
        for p in self.populations.values():
            first_gid += p.POP_SIZE

        # create NetworkPopulation object
        population = NetworkPopulation(CWD=CWD, CELLPATH=CELLPATH, first_gid=first_gid,
                                Cell=Cell,
                                POP_SIZE=POP_SIZE, name=name,
                                cell_args=cell_args, pop_args=pop_args,
                                rotation_args=rotation_args,
                                OUTPUTPATH=self.OUTPUTPATH)

        # associate gids of cells on this RANK such that NEURON can look up
        # at which RANK different cells are created when connecting the network
        for gid in population.gids:
            self.pc.set_gid2node(gid, RANK)

        # Prepare connection targets by iterating over local neurons in pop.
        for gid, cell in zip(population.gids, population.cells):
            # attach NetCon source (spike detektor) to each cell's soma with no
            # target to cell gid
            cell.create_spike_detector(None)
            # assosiate cell gid with the NetCon source
            self.pc.cell(gid, cell.sd_netconlist[-1])

            # record spike events
            population.spike_vectors.append(neuron.h.Vector())
            cell.sd_netconlist[-1].record(population.spike_vectors[-1])

        # add population object to dictionary of populations
        self.populations[name] = population

        # append population name to list (Network.populations.keys() not unique)
        self.population_names.append(name)


    def get_connectivity_rand(self, pre='L5PC', post='L5PC', connprob = 0.2):
        """
        Dummy function creating a (boolean) cell to cell connectivity matrix
        between pre and postsynaptic populations.

        Connections are drawn randomly between presynaptic cell gids in
        population 'pre' and postsynaptic cell gids in 'post' on this RANK with
        a fixed connection probability. self-connections are disabled if
        presynaptic and postsynaptic populations are the same.

        Parameters
        ----------
        pre : str
            presynaptic population name
        post : str
            postsynaptic population name
        connprob : float in [0, 1]
            connection probability, connections are drawn on random

        Returns
        -------
        ndarray, dtype bool
            n_pre x n_post array of connections between n_pre presynaptic
            neurons and n_post postsynaptic neurons on this RANK. Entries
            with True denotes a connection.
        """
        n_pre = self.populations[pre].POP_SIZE
        gids = np.array(self.populations[post].gids).astype(int)

        # first check if there are any postsyn cells on this RANK
        if gids.size > 0:
            # define incoming connections for cells on this RANK
            C = np.random.rand(n_pre, gids.size) < connprob
            if pre == post:
                # avoid self connections.
                gids_pre, gids_post = np.where(C)
                gids_pre += self.populations[pre].first_gid
                gids_post *= SIZE # asssume round-robin distribution of gids
                gids_post += self.populations[post].gids[0]
                inds = gids_pre == gids_post
                gids_pre = gids_pre[inds == False]
                gids_pre -= self.populations[pre].first_gid
                gids_post = gids_post[inds == False]
                gids_post -= self.populations[post].gids[0]
                gids_post //= SIZE
                c = np.c_[gids_pre, gids_post]
                # create boolean matrix
                C = ss.csr_matrix((np.ones(gids_pre.shape[0], dtype=bool),
                                   (c[:, 0], c[:, 1])),
                                  shape=(n_pre, gids.size), dtype=bool)
                return C.toarray()
            else:
                return C
        else:
            return np.zeros((n_pre, 0), dtype=bool)


    def connect(self, pre, post, connectivity,
                syntype=neuron.h.ExpSyn,
                synparams=dict(tau=2., e=0.),
                weightfun=np.random.normal,
                weightargs=dict(loc=0.1, scale=0.01),
                minweight=0,
                delayfun=np.random.normal,
                delayargs=dict(loc=2, scale=0.2),
                mindelay=0.3,
                multapsefun=np.random.normal,
                multapseargs=dict(loc=4, scale=1),
                syn_pos_args=dict(section=['soma', 'dend', 'apic'],
                                  fun=[stats.norm]*2,
                                  funargs=[dict(loc=0, scale=100)]*2,
                                  funweights=[0.5]*2,
                                  z_min=-1E6, z_max=1E6,
                                  ),
                save_connections=False,
                ):
        """
        Connect presynaptic cells to postsynaptic cells. Connections are
        drawn from presynaptic cells to postsynaptic cells, hence connectivity
        array must only be specified for postsynaptic units existing on this
        RANK.

        Parameters
        ----------
        pre : str
            presynaptic population name
        post : str
            postsynaptic population name
        connectivity : ndarray / (scipy.sparse array)
            boolean connectivity matrix between pre and post.
        syntype : hoc.HocObject
            reference to NEURON synapse mechanism, e.g., neuron.h.ExpSyn
        synparams : dict
            dictionary of parameters for synapse mechanism, keys 'e', 'tau' etc.
        weightfun : function
            function used to draw weights from a numpy.random distribution
        weightargs : dict
            parameters passed to weightfun
        minweight : float,
            minimum weight in units of nS
        delayfun : function
            function used to draw delays from a numpy.random distribution
        delayargs : dict
            parameters passed to delayfun
        mindelay : float,
            minimum delay in multiples of dt
        multapsefun : function or None
            function reference, e.g., numpy.random.normal used to draw a number
            of synapses for a cell-to-cell connection. If None, draw only one
            connection
        multapseargs : dict
            arguments passed to multapsefun
        syn_pos_args : dict
            arguments passed to inherited LFPy.Cell method
            NetworkCell.get_rand_idx_area_and_distribution_norm to find
            synapse locations.
        save_connections : bool
            if True (default False), save instantiated connections to HDF5 file
            "Network.OUTPUTPATH/synapse_positions.h5" as dataset "<pre>:<post>"
            using a structured ndarray with dtype
            [('gid', 'i8'), ('x', float), ('y', float), ('z', float)]
            where gid is postsynaptic cell id, and x,y,z the corresponding
            midpoint coordinates of the target compartment.
        """
        # set up connections from all cells in presynaptic to post across RANKs
        n0 = self.populations[pre].first_gid
        # gids of presynaptic neurons:
        pre_gids = np.arange(n0, n0 + self.populations[pre].POP_SIZE)

        # count connections and synapses made on this RANK
        conncount = connectivity.astype(int).sum()
        syncount = 0

        # keep track of synapse positions for this connect
        # call on this rank such that these can be communicated and stored
        syn_idx_pos = []

        # iterate over gids on this RANK and create connections
        for i, (post_gid, cell) in enumerate(zip(self.populations[post].gids, self.populations[post].cells)):
            # do NOT iterate over all possible presynaptic neurons
            for pre_gid in pre_gids[connectivity[:, i]]:
                # throw a warning if sender neuron is identical to receiving neuron
                if post_gid == pre_gid:
                    print('connecting cell w. gid {} to itself (RANK {})'.format(post_gid, RANK))

                # assess number of synapses
                if multapsefun is None:
                    nidx = 1
                else:
                    nidx = 0
                    j = 0
                    while nidx <= 0 and j < 1000:
                        nidx = int(multapsefun(**multapseargs))
                        j += 1
                    if j == 1000:
                        raise Exception('change multapseargs as no positive synapse count was found in 1000 trials')

                # find synapse locations and corresponding section names
                idxs = cell.get_rand_idx_area_and_distribution_norm(nidx=nidx, **syn_pos_args)
                secs = cell.get_idx_name(idxs)

                # draw weights
                weights = weightfun(size=nidx, **weightargs)
                # redraw weights less that minweight
                while np.any(weights < minweight):
                    j = weights < minweight
                    weights[j] = weightfun(size=j.sum(), **weightargs)

                # draw delays
                delays = delayfun(size=nidx, **delayargs)
                # redraw delays shorter than mindelay
                while np.any(delays < mindelay):
                    j = delays < mindelay
                    delays[j] = delayfun(size=j.sum(), **delayargs)

                for i, ((idx, secname, x), weight, delay) in enumerate(zip(secs, weights, delays)):
                    cell.create_synapse(cell,
                                        # TODO: Find neater way of accessing Section reference, this seems slow
                                        sec=list(cell.allseclist)[np.where(np.array(cell.allsecnames)==secname)[0][0]],
                                        x=x, syntype=syntype,
                                        synparams=synparams)
                    # connect up NetCon object
                    nc = self.pc.gid_connect(pre_gid, cell.netconsynapses[-1])
                    nc.weight[0] = weight
                    nc.delay = delays[i]
                    self.netconlist.append(nc)

                    # store also synapse indices allowing for computing LFPs from syn.i
                    cell.synidx.append(idx)

                    # store gid and xyz-coordinate of synapse positions
                    syn_idx_pos.append((cell.gid, cell.xmid[idx], cell.ymid[idx], cell.zmid[idx]))

                syncount += nidx

        conncount = COMM.reduce(conncount, op=MPI.SUM, root=0)
        syncount = COMM.reduce(syncount, op=MPI.SUM, root=0)

        if RANK == 0:
            print('Connected population {} to {} by {} connections and {} synapses'.format(pre, post, conncount, syncount))

        else:
            conncount = None
            syncount = None


        # gather and write syn_idx_pos data
        if save_connections:
            if RANK == 0:
                synData = flattenlist(COMM.gather(syn_idx_pos))

                # convert to structured array
                dtype = [('gid', 'i8'), ('x', float), ('y', float), ('z', float)]
                synDataArray = np.empty((len(synData), ), dtype=dtype)
                for i, (gid, x, y, z) in enumerate(synData):
                    synDataArray[i]['gid'] = gid
                    synDataArray[i]['x'] = x
                    synDataArray[i]['y'] = y
                    synDataArray[i]['z'] = z
                # Dump to hdf5 file, append to file if entry exists
                f = h5py.File(os.path.join(self.OUTPUTPATH,
                                           'synapse_positions.h5'), 'a')
                key = '{}:{}'.format(pre, post)
                if key in f.keys():
                    del f[key]
                    try:
                        assert key not in f.keys()
                    except AssertionError:
                        raise AssertionError
                f[key] = synDataArray
                f.close()
            else:
                COMM.gather(syn_idx_pos)

        return COMM.bcast([conncount, syncount])

    def enable_extracellular_stimulation(self, electrode, t_ext=None, n=1, seed=None):
        raise NotImplementedError()

    def simulate(self, electrode=None, rec_imem=False, rec_vmem=False,
                 rec_ipas=False, rec_icap=False,
                 rec_isyn=False, rec_vmemsyn=False, rec_istim=False,
                 rec_current_dipole_moment=False,
                 rec_pop_contributions=False,
                 rec_variables=[], variable_dt=False, atol=0.001,
                 to_memory=True, to_file=False,
                 file_name='OUTPUT.h5',
                 dotprodcoeffs=None, **kwargs):
        """
        This is the main function running the simulation of the network model.

        Parameters
        ----------
        electrode:
            Either an LFPy.RecExtElectrode object or a list of such.
                    If supplied, LFPs will be calculated at every time step
                    and accessible as electrode.LFP. If a list of objects
                    is given, accessible as electrode[0].LFP etc.
        rec_imem:   If true, segment membrane currents will be recorded
                    If no electrode argument is given, it is necessary to
                    set rec_imem=True in order to calculate LFP later on.
                    Units of (nA).
        rec_vmem:   record segment membrane voltages (mV)
        rec_ipas:   record passive segment membrane currents (nA)
        rec_icap:   record capacitive segment membrane currents (nA)
        rec_isyn:   record synaptic currents of from Synapse class (nA)
        rec_vmemsyn:    record membrane voltage of segments with Synapse(mV)
        rec_istim:  record currents of StimIntraElectrode (nA)
        rec_current_dipole_moment : bool
            If True, compute and record current-dipole moment from
            transmembrane currents as in Linden et al. (2010) J Comput Neurosci,
            DOI: 10.1007/s10827-010-0245-4. Will set the `LFPy.Cell` attribute
            `current_dipole_moment` as n_timesteps x 3 `ndarray` where the
            last dimension contains the x,y,z components of the dipole moment.
        rec_pop_contributions : bool
            If True, compute and return single-population contributions to
            the extracellular potential during simulation time
        rec_variables: list of variables to record, i.e arg=['cai', ]
        variable_dt: boolean, using variable timestep in NEURON
        atol:       absolute tolerance used with NEURON variable timestep
        to_memory:  only valid with electrode, store lfp in -> electrode.LFP
        to_file:    only valid with electrode, save LFPs in hdf5 file format
        file_name : str
            If to_file is True, file which extracellular potentials will be
            written to. The file format is HDF5, default is "OUTPUT.h5", put
            in folder Network.OUTPUTPATH
        dotprodcoeffs :  list of N x Nseg ndarray. These arrays will at
                    every timestep be multiplied by the membrane currents.
                    Presumably useful for memory efficient csd or lfp calcs
        **kwargs :  keyword argument dict values passed along to function
                    _run_simulation_with_electrode(), containing some or all of
                    the boolean flags: use_ipas, use_icap, use_isyn
                    (defaulting to 'False').

        Returns
        -------
        SPIKES : dict
            the first returned argument is a dictionary with keys 'gids' and
            'times'. Each item is a nested list of len(Npop) times N_X where N_X
            is the corresponding population size. Each entry is a np.ndarray
            containing the spike times of each cell in the nested list in item
            'gids'
        OUTPUT : list of ndarray
            if parameters electrode is not None and/or dotprodcoeffs is not
            None, contains the
            [electrode.LFP, ...., (dotprodcoeffs[0] dot I)(t), ...]
            The first output is a structured array, so OUTPUT[0]['imem']
            corresponds to the total LFP and the other the per-population
            contributions.
        P : ndarray
            if rec_current_dipole_moment==True, contains the x,y,z-components of
            current-dipole moment from transmembrane currents summed up over
            all populations

        """
        # set up integrator, use the CVode().fast_imem method by default
        # as it doesn't hurt sim speeds much if at all.
        cvode = neuron.h.CVode()
        try:
            cvode.use_fast_imem(1)
        except AttributeError as ae:
            raise Exception('neuron.h.CVode().use_fast_imem() not found. Please update NEURON to v.7.4 or newer')

        # test some of the inputs
        try:
            if electrode is None:
                assert(rec_pop_contributions is False)
        except AssertionError:
            raise AssertionError('rec_pop_contributions can not be True when electrode is None')

        if not variable_dt:
            dt = self.dt
        else:
            dt = None

        for name in self.population_names:
            for cell in self.populations[name].cells:
                cell._set_soma_volt_recorder(dt)
                if rec_imem:
                    cell._set_imem_recorders(dt)
                if rec_vmem:
                    cell._set_voltage_recorders(dt)
                if rec_ipas:
                    cell._set_ipas_recorders(dt)
                if rec_icap:
                    cell._set_icap_recorders(dt)
                # if rec_current_dipole_moment:
                #     self._set_current_dipole_moment_array()
                if len(rec_variables) > 0:
                    cell._set_variable_recorders(rec_variables)

        #run fadvance until t >= tstop, and calculate LFP if asked for
        if electrode is None and dotprodcoeffs is None and not rec_current_dipole_moment and not rec_pop_contributions and not to_file:
            if not rec_imem:
                if self.verbose:
                    print("rec_imem = {}, not recording membrane currents!".format(rec_imem))
            _run_simulation(self, cvode, variable_dt, atol)
        else:
            if dotprodcoeffs is not None:
                raise NotImplementedError
            LFP, P = _run_simulation_with_electrode(self, cvode=cvode,
                            electrode=electrode,
                            variable_dt=variable_dt,
                            atol=atol,
                            to_memory=to_memory,
                            to_file=to_file,
                            file_name='tmp_output_RANK_{:03d}.h5',
                            dotprodcoeffs=dotprodcoeffs,
                            rec_current_dipole_moment=rec_current_dipole_moment,
                            rec_pop_contributions=rec_pop_contributions,
                            **kwargs)

        for name in self.population_names:
            for cell in self.populations[name].cells:
                #somatic trace
                cell.somav = np.array(cell.somav)
                if rec_imem:
                    cell._calc_imem()
                if rec_ipas:
                    cell._calc_ipas()
                if rec_icap:
                    cell._calc_icap()
                if rec_vmem:
                    cell._collect_vmem()
                if rec_isyn:
                    cell._collect_isyn()
                if rec_vmemsyn:
                    cell._collect_vsyn()
                if rec_istim:
                    cell._collect_istim()
                if len(rec_variables) > 0:
                    cell._collect_rec_variables(rec_variables)
                if hasattr(cell, 'netstimlist'):
                    del cell.netstimlist

        # Collect spike trains across all RANKs to RANK 0
        for name in self.population_names:
            population = self.populations[name]
            for i in range(len(population.spike_vectors)):
                population.spike_vectors[i] = np.array(population.spike_vectors[i])
        if RANK == 0:
            times = []
            gids = []
            for i, name in enumerate(self.population_names):
                times.append([])
                gids.append([])
                times[i] += [x for x in self.populations[name].spike_vectors]
                gids[i] += [x for x in self.populations[name].gids]
                for j in range(1, SIZE):
                    times[i] += COMM.recv(source=j, tag=13)
                    gids[i] += COMM.recv(source=j, tag=14)
        else:
            times = None
            gids = None
            for name in self.population_names:
                COMM.send([x for x in self.populations[name].spike_vectors],
                    dest=0, tag=13)
                COMM.send([x for x in self.populations[name].gids],
                    dest=0, tag=14)

        # create final output file, summing up single RANK output from temp files
        if to_file and electrode is not None:
            op=MPI.SUM
            fname = os.path.join(self.OUTPUTPATH, 'tmp_output_RANK_{:03d}.h5'.format(RANK))
            f0 = h5py.File(fname, 'r')
            if RANK == 0:
                f1 = h5py.File(os.path.join(self.OUTPUTPATH, file_name), 'w')
            dtype = []
            for key, value in f0[list(f0.keys())[0]].items():
                dtype.append((str(key), np.float))
            shape = value.shape
            for grp in f0.keys():
                if RANK == 0:
                    f1[grp] = np.zeros(shape, dtype=dtype)
                for key, value in f0[grp].items():
                    if RANK == 0:
                        recvbuf = np.zeros(shape, dtype=np.float)
                    else:
                        recvbuf = None
                    COMM.Reduce(value[()].astype(np.float), recvbuf, op=op, root=0)
                    if RANK == 0:
                        f1[grp][key] = recvbuf
            f0.close()
            if RANK == 0:
                f1.close()
            os.remove(fname)


        if electrode is None and dotprodcoeffs is None and not rec_current_dipole_moment and not rec_pop_contributions:
            return dict(times=times, gids=gids)
        else:
            # communicate and sum up LFPs and dipole moments:
            if LFP is not None:
                for i in range(len(LFP)):
                    LFP[i] = ReduceStructArray(LFP[i])
            if P is not None:
                P = ReduceStructArray(P)
            return dict(times=times, gids=gids), LFP, P


    def _create_network_dummycell(self):
        """
        set up parameters for a DummyCell object, allowing for computing
        the sum of all single-cell LFPs at each timestep, essentially
        creating one supercell with all segments of all cell objects
        present on this RANK.
        """
        # compute the total number of segments per population on this RANK
        nsegs = [[cell.totnsegs for cell in self.populations[name].cells]
            for name in self.population_names]
        for i, nseg in enumerate(nsegs):
            if nseg == []:
                nsegs[i] = [0]
        for i, y in enumerate(nsegs): nsegs[i] = np.sum(y)
        nsegs = np.array(nsegs, dtype=int)

        totnsegs = nsegs.sum()
        imem = np.eye(totnsegs)
        xstart = np.array([])
        xmid = np.array([])
        xend = np.array([])
        ystart = np.array([])
        ymid = np.array([])
        yend = np.array([])
        zstart = np.array([])
        zmid = np.array([])
        zend = np.array([])
        diam = np.array([])
        area = np.array([])

        somainds = np.array([], dtype=int)
        nseg = 0

        for name in self.population_names:
            for cell in self.populations[name].cells:
                xstart = np.r_[xstart, cell.xstart]
                ystart = np.r_[ystart, cell.ystart]
                zstart = np.r_[zstart, cell.zstart]
                xmid = np.r_[xmid, cell.xmid]
                ymid = np.r_[ymid, cell.ymid]
                zmid = np.r_[zmid, cell.zmid]
                xend = np.r_[xend, cell.xend]
                yend = np.r_[yend, cell.yend]
                zend = np.r_[zend, cell.zend]
                diam = np.r_[diam, cell.diam]
                area = np.r_[area, cell.area]

                somainds = np.r_[somainds, cell.get_idx("soma")+nseg]
                nseg += cell.totnsegs


        # return number of segments per population and DummyCell object
        return nsegs, DummyCell(totnsegs,
                         imem,
                         xstart, xmid, xend,
                         ystart, ymid, yend,
                         zstart, zmid, zend,
                         diam, area, somainds)


def _run_simulation(network, cvode, variable_dt=False, atol=0.001):
    """
    Running the actual simulation in NEURON, simulations in NEURON
    are now interruptable.

    Parameters
    ----------
    network : LFPy.Network object
        instantiation of class LFPy.Network
    cvode : neuron.h.CVode() object
    variable_dt : bool
        switch for variable-timestep method
    atol : float
        absolute tolerance with CVode for variable time-step method
    """
    # set maximum integration step, it is necessary for communication of
    # spikes across RANKs to occur.
    network.pc.set_maxstep(10)

    # time resolution
    neuron.h.dt = network.dt

    # needed for variable dt method
    if variable_dt:
        cvode.active(1)
        cvode.atol(atol)
    else:
        cvode.active(0)

    # initialize state
    neuron.h.finitialize(network.v_init)

    # initialize current- and record
    if cvode.active():
        cvode.re_init()
    else:
        neuron.h.fcurrent()
    neuron.h.frecord_init()

    # Starting simulation at tstart
    neuron.h.t = network.tstart

    # only needed if LFPy.Synapse classes are used.
    for name in network.population_names:
        for cell in network.populations[name].cells:
            cell._loadspikes()

    while neuron.h.t < network.tstop:
        neuron.h.fadvance()
        if neuron.h.t % 100 == 0:
            if RANK == 0:
                print('t = {} ms'.format(neuron.h.t))

    return


def _run_simulation_with_electrode(network, cvode,
                                   electrode=None,
                                   variable_dt=False,
                                   atol=0.001,
                                   to_memory=True,
                                   to_file=False,
                                   file_name=None,
                                   dotprodcoeffs=None,
                                   rec_current_dipole_moment=False,
                                   use_ipas=False, use_icap=False,
                                   use_isyn=False,
                                   rec_pop_contributions=False
                                   ):
    """
    Running the actual simulation in NEURON.
    electrode argument used to determine coefficient
    matrix, and calculate the LFP on every time step.

    Parameters
    ----------
    network : LFPy.Network object
        instantiation of class LFPy.Network
    cvode : neuron.h.CVode() object
    electrode : LFPy.RecExtElectrode object or None
        instantiation of class LFPy.RecExtElectrode for which extracellular
        potentials will be computed.
    variable_dt : bool
        switch for variable-timestep method
    atol : float
        absolute tolerance with CVode for variable time-step method
    to_memory : bool
        Boolean flag for computing extracellular potentials, default is True
    to_file : bool or None
        Boolean flag for computing extracellular potentials to file
        <OUTPUTPATH/file_name>, default is False
    file_name : formattable str
        If to_file is True, file which extracellular potentials will be
        written to. The file format is HDF5, default is
        "output_RANK_{:03d}.h5". The output is written per RANK, and the
        RANK # will be inserted into the corresponding file name.
    dotprodcoeffs : None or list of ndarrays
        Each element in list is a mapping of transmembrane currents to a measure
        on the form :math:`V = \\mathbf{C} \\cdot \\mathbf{I}`
    rec_current_dipole_moment : bool
        if True, compute and store the total current-dipole moment per time
        step as the sum over each individual population
    use_ipas : bool
        if True, compute the contribution to extracellular potentials across
        the passive leak channels embedded in the cells membranes summed over
        populations
    use_icap : bool
        if True, compute the contribution to extracellular potentials across
        the membrane capacitance embedded in the cells membranes summed over
        populations
    use_isyn : bool
        if True, compute the contribution to extracellular potentials across
        the excitatory and inhibitory synapses embedded in the cells membranes
        summed over populations
    rec_pop_contributions : bool
        if True, compute and return single-population contributions to the
        extracellular potential during each time step of the simulation

    Returns
    -------
    RESULTS : list
        ordered according to [dotprodcoeffs, ..., electrode, ...], each element
        being the superimposed contribution to i.e., the extracellular potential
        at each timestep from all cell objects on this particular RANK.
        Thus, no single-cell contributions to the LFP
        are returned.
    DIPOLE_MOMENT : ndarray
        Shape (n_timesteps, 3) array containing the x,y,z-components of the
        current-dipole moment summed up over contributions from cells across
        all populations on this MPI RANK.
    """
    # create a dummycell object lumping together needed attributes
    # for calculation of extracellular potentials etc. The population_nsegs
    # array is used to slice indices such that single-population
    # contributions to the potential can be calculated.
    population_nsegs, network_dummycell = network._create_network_dummycell()

    # Use electrode object(s) to calculate coefficient matrices for LFP
    # calculations. If electrode is a list, then
    # put electrodecoeff in a list, if it isn't already
    if dotprodcoeffs is not None:
        if type(dotprodcoeffs) != list:
            dotprodcoeffs = [dotprodcoeffs]
    else:
        #create empty list if no dotprodcoeffs are supplied
        dotprodcoeffs = []

    #access electrode object and append dotprodcoeffs
    if electrode is not None:
        #put electrode argument in list if needed
        if type(electrode) == list:
            electrodes = electrode
        else:
            electrodes = [electrode]

        # At each timestep we will later construct a single vector I of all
        # transmembrane currents. With that, and a corresponding matrix G
        # mapping a current contribution to an electrode contact, we can here
        # compute the extracellular potentials V_r in all contacts r at
        # timestep t_i as
        # V_r(r, t_i) = G x I(r, t_i)


        # # create a dummycell object lumping together needed attributes
        # # for calculation of extracellular potentials. The population_nsegs
        # # array is used to slice indices such that single-population
        # # contributions to the potential can be calculated.
        # population_nsegs, network_dummycell = network._create_network_dummycell()

        # We can have a number of separate electrode objects in a list, create
        # mappings for each
        for el in electrodes:
            # el.calc_lfp(cell=network_dummycell)
            el.calc_mapping(cell=network_dummycell)
            dotprodcoeffs += [el.mapping]
            # del el.LFP
            del el.mapping

    elif electrode is None:
        electrodes = None
        # if rec_current_dipole_moment:
        #     population_nsegs, network_dummycell = network._create_network_dummycell()




    # set maximum integration step, it is necessary for communication of
    # spikes across RANKs to occur.
    # NOTE: Should this depend on the minimum delay in the network?
    network.pc.set_maxstep(10)

    # Initialize NEURON simulations of cell object
    neuron.h.dt = network.dt

    # needed for variable dt method
    if variable_dt:
        cvode.active(1)
        cvode.atol(atol)
    else:
        cvode.active(0)

    #initialize state
    neuron.h.finitialize(network.v_init)

    # use fast calculation of transmembrane currents
    cvode.use_fast_imem(1)

    #initialize current- and record
    if cvode.active():
        cvode.re_init()
    else:
        neuron.h.fcurrent()
    neuron.h.frecord_init()

    #Starting simulation at tstart
    neuron.h.t = network.tstart

    # create list of cells across all populations to simplify loops
    cells = []
    for name in network.population_names:
        cells += network.populations[name].cells

    #load spike times from NetCon, only needed if LFPy.Synapse class is used
    for cell in cells:
        cell._loadspikes()


    # define data type for structured arrays dependent on the boolean arguments
    dtype = [('imem', np.float)]
    if use_ipas: dtype += [('ipas', np.float)]
    if use_icap: dtype += [('icap', np.float)]
    if use_isyn: dtype += [('isyn_e', np.float), ('isyn_i', np.float)]
    if rec_pop_contributions: dtype += list(zip(network.population_names,
                                                [np.float]*len(network.population_names)))

    # setup list of structured arrays for all extracellular potentials
    # at each contact from different source terms and subpopulations
    if to_memory:
        RESULTS = []
        for coeffs in dotprodcoeffs:
            RESULTS.append(np.zeros((coeffs.shape[0],
                                     int(network.tstop / network.dt) + 1),
                                    dtype=dtype)
                           )
    else:
        RESULTS = None

    # container for electric current dipole moment for the individual
    # populations captured inside the DummyCell instance
    if rec_current_dipole_moment:
        DIPOLE_MOMENT = np.zeros((int(network.tstop / network.dt) + 1, 3),
            dtype=list(zip(network.population_names,
                           [np.float]*len(network.population_names))))
    else:
        DIPOLE_MOMENT = None

    #LFPs for each electrode will be put here during simulations
    if to_file:
        #ensure right ending:
        if file_name.split('.')[-1] != 'h5':
            file_name += '.h5'
        outputfile = h5py.File(os.path.join(network.OUTPUTPATH,
                                            file_name.format(RANK)), 'w')
        for i, coeffs in enumerate(dotprodcoeffs):
            # can't do it this way until h5py issue #770
            # (https://github.com/h5py/h5py/issues/770) is fixed:
            # outputfile['OUTPUT[{}]'.format(i)] = np.zeros((coeffs.shape[0],
            #                     int(network.tstop / network.dt) + 1), dtype=dtype)
            grp = outputfile.create_group('OUTPUT[{}]'.format(i))
            for key, val in dtype:
                grp[key] = np.zeros((coeffs.shape[0], int(network.tstop / network.dt) + 1), dtype=val)

    # temp vector to store membrane currents at each timestep:
    imem = np.zeros(network_dummycell.totnsegs, dtype=dtype)

    # create a 2D array representation of segment midpoints for dot product
    # with transmembrane currents when computing dipole moment
    if rec_current_dipole_moment:
        midpoints = np.c_[network_dummycell.xmid,
                          network_dummycell.ymid,
                          network_dummycell.zmid]

    #run fadvance until time limit, and calculate LFPs for each timestep
    tstep = 0
    while neuron.h.t < network.tstop:
        if neuron.h.t >= 0:
            i = 0
            totnsegs = 0
            if use_isyn:
                imem['isyn_e'] = 0. # need to reset these for every iteration
                imem['isyn_i'] = 0. # because we sum over synapses
            for cell in cells:
                for sec in cell.allseclist:
                    for seg in sec:
                        imem['imem'][i] = seg.i_membrane_
                        if use_ipas:
                            imem['ipas'][i] = seg.i_pas
                        if use_icap:
                            imem['icap'][i] = seg.i_cap
                        i += 1

                if use_isyn:
                    for idx, syn in zip(cell.synidx, cell.netconsynapses):
                        if hasattr(syn, 'e') and syn.e > -50:
                            imem['isyn_e'][idx+totnsegs] += syn.i
                        else:
                            imem['isyn_i'][idx+totnsegs] += syn.i

                totnsegs += cell.totnsegs

            if rec_current_dipole_moment:
                k = 0 # counter
                for nsegs, name in zip(population_nsegs, network.population_names):
                    cellinds = np.arange(k, k+nsegs)
                    DIPOLE_MOMENT[name][tstep, ] = np.dot(imem['imem'][cellinds, ],
                                                          midpoints[cellinds, ])
                    k += nsegs

            if to_memory:
                for j, coeffs in enumerate(dotprodcoeffs):
                    RESULTS[j]['imem'][:, tstep] = np.dot(coeffs, imem['imem'])
                    if use_ipas:
                        RESULTS[j]['ipas'][:, tstep] = np.dot(coeffs, imem['ipas'] * network_dummycell.area * 1E-2)
                    if use_icap:
                        RESULTS[j]['icap'][:, tstep] = np.dot(coeffs, imem['icap'] * network_dummycell.area * 1E-2)
                    if use_isyn:
                        RESULTS[j]['isyn_e'][:, tstep] = np.dot(coeffs, imem['isyn_e'])
                        RESULTS[j]['isyn_i'][:, tstep] = np.dot(coeffs, imem['isyn_i'])

                if rec_pop_contributions:
                    for j, coeffs in enumerate(dotprodcoeffs):
                        k = 0 # counter
                        for nsegs, name in zip(population_nsegs, network.population_names):
                            cellinds = np.arange(k, k+nsegs)
                            RESULTS[j][name][:, tstep] = np.dot(coeffs[:, cellinds],
                                                                imem['imem'][cellinds, ])
                            k += nsegs

            if to_file:
                for j, coeffs in enumerate(dotprodcoeffs):
                    outputfile['OUTPUT[{}]'.format(j)
                               ]['imem'][:, tstep] = np.dot(coeffs, imem['imem'])
                    if use_ipas:
                        outputfile['OUTPUT[{}]'.format(j)
                                   ]['ipas'][:, tstep] = np.dot(coeffs, imem['ipas'] * network_dummycell.area * 1E-2)
                    if use_icap:
                        outputfile['OUTPUT[{}]'.format(j)
                                   ]['icap'][:, tstep] = np.dot(coeffs, imem['icap'] * network_dummycell.area * 1E-2)
                    if use_isyn:
                        outputfile['OUTPUT[{}]'.format(j)
                                   ]['isyn_e'][:, tstep] = np.dot(coeffs, imem['isyn_e'])
                        outputfile['OUTPUT[{}]'.format(j)
                                   ]['isyn_i'][:, tstep] = np.dot(coeffs, imem['isyn_i'])

                if rec_pop_contributions:
                    for j, coeffs in enumerate(dotprodcoeffs):
                        k = 0 # counter
                        for nsegs, name in zip(population_nsegs, network.population_names):
                            cellinds = np.arange(k, k+nsegs)
                            outputfile['OUTPUT[{}]'.format(j)
                                       ][name][:, tstep] = np.dot(coeffs[:, cellinds], imem['imem'][cellinds, ])
                            k += nsegs

            tstep += 1
        neuron.h.fadvance()
        if neuron.h.t % 100. == 0.:
            if RANK == 0:
                print('t = {} ms'.format(neuron.h.t))


    try:
        #calculate LFP after final fadvance()
        i = 0
        totnsegs = 0
        if use_isyn:
            imem['isyn_e'] = 0. # need to reset these for every iteration because we sum over synapses
            imem['isyn_i'] = 0.
        for cell in cells:
            for sec in cell.allseclist:
                for seg in sec:
                    imem['imem'][i] = seg.i_membrane_
                    if use_ipas:
                        imem['ipas'][i] = seg.i_pas
                    if use_icap:
                        imem['icap'][i] = seg.i_cap
                    i += 1

            if use_isyn:
                for idx, syn in zip(cell.synidx, cell.netconsynapses):
                    if hasattr(syn, 'e') and syn.e > -50:
                        imem['isyn_e'][idx+totnsegs] += syn.i
                    else:
                        imem['isyn_i'][idx+totnsegs] += syn.i

            totnsegs += cell.totnsegs

        if rec_current_dipole_moment:
            k = 0 # counter
            for nsegs, name in zip(population_nsegs, network.population_names):
                cellinds = np.arange(k, k+nsegs)
                DIPOLE_MOMENT[name][tstep, ] = np.dot(imem['imem'][cellinds, ], midpoints[cellinds, ])
                k += nsegs

        if to_memory:
            for j, coeffs in enumerate(dotprodcoeffs):
                RESULTS[j]['imem'][:, tstep] = np.dot(coeffs, imem['imem'])
                if use_ipas:
                    RESULTS[j]['ipas'][:, tstep] = np.dot(coeffs, imem['ipas'] * network_dummycell.area * 1E-2)
                if use_icap:
                    RESULTS[j]['icap'][:, tstep] = np.dot(coeffs, imem['icap'] * network_dummycell.area * 1E-2)
                if use_isyn:
                    RESULTS[j]['isyn_e'][:, tstep] = np.dot(coeffs, imem['isyn_e'])
                    RESULTS[j]['isyn_i'][:, tstep] = np.dot(coeffs, imem['isyn_i'])

            if rec_pop_contributions:
                for j, coeffs in enumerate(dotprodcoeffs):
                    k = 0 # counter
                    for nsegs, name in zip(population_nsegs, network.population_names):
                        cellinds = np.arange(k, k+nsegs)
                        RESULTS[j][name][:, tstep] = np.dot(coeffs[:, cellinds], imem['imem'][cellinds, ])
                        k += nsegs


        if to_file:
            for j, coeffs in enumerate(dotprodcoeffs):
                outputfile['OUTPUT[{}]'.format(j)
                           ]['imem'][:, tstep] = np.dot(coeffs, imem['imem'])
                if use_ipas:
                    outputfile['OUTPUT[{}]'.format(j)
                               ]['ipas'][:, tstep] = np.dot(coeffs, imem['ipas'] * network_dummycell.area * 1E-2)
                if use_icap:
                    outputfile['OUTPUT[{}]'.format(j)
                               ]['icap'][:, tstep] = np.dot(coeffs, imem['icap'] * network_dummycell.area * 1E-2)
                if use_isyn:
                    outputfile['OUTPUT[{}]'.format(j)
                               ]['isyn_e'][:, tstep] = np.dot(coeffs, imem['isyn_e'])
                    outputfile['OUTPUT[{}]'.format(j)
                               ]['isyn_i'][:, tstep] = np.dot(coeffs, imem['isyn_i'])

            if rec_pop_contributions:
                for j, coeffs in enumerate(dotprodcoeffs):
                    k = 0 # counter
                    for nsegs, name in zip(population_nsegs, network.population_names):
                        cellinds = np.arange(k, k+nsegs)
                        outputfile['OUTPUT[{}]'.format(j)
                                   ][name][:, tstep] = np.dot(coeffs[:, cellinds],
                                                              imem['imem'][cellinds, ])
                        k += nsegs

    except IndexError:
        pass

    if to_memory:
        return RESULTS, DIPOLE_MOMENT

    if to_file:
        outputfile.close()
        return RESULTS, DIPOLE_MOMENT


def ReduceStructArray(sendbuf, op=MPI.SUM):
    """
    simplify MPI Reduce for structured ndarrays with floating point numbers

    Parameters
    ----------
    sendbuf : structured ndarray
        Array data to be reduced (default: summed)
    op : mpi4py.MPI.Op object
        MPI_Reduce function. Default is mpi4py.MPI.SUM
    """
    if RANK == 0:
        shape = sendbuf.shape
        dtype_names = sendbuf.dtype.names
    else:
        shape = None
        dtype_names = None
    shape = COMM.bcast(shape)
    dtype_names = COMM.bcast(dtype_names)

    if RANK == 0:
        reduced = np.zeros(shape,
                           dtype=list(zip(dtype_names,
                                     ['f8' for i in range(len(dtype_names))])))
    else:
        reduced = None
    for name in dtype_names:
        if RANK == 0:
            recvbuf = np.zeros(shape)
        else:
            recvbuf = None
        COMM.Reduce(np.array(sendbuf[name]), recvbuf, op=op, root=0)
        if RANK == 0:
            reduced[name] = recvbuf
    return reduced



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# cython: language_level=2
"""Copyright (C) 2012 Computational Neuroscience Group, NMBU.

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

"""


from time import time
import numpy as np
cimport numpy as np
import neuron
from pathlib import Path

DTYPE = np.float64
ctypedef np.float64_t DTYPE_t
ctypedef Py_ssize_t   LTYPE_t


def _run_simulation(cell, cvode, variable_dt=False, atol=0.001):
    """
    Running the actual simulation in NEURON, simulations in NEURON
    is now interruptable.
    """
    neuron.h.dt = cell.dt

    # variable dt method
    if variable_dt:
        cvode.active(1)
        cvode.atol(atol)
    else:
        cvode.active(0)

    #re-initialize state
    neuron.h.finitialize(cell.v_init)

    #initialize current- and record
    if cvode.active():
        cvode.re_init()
    else:
        neuron.h.fcurrent()
    neuron.h.frecord_init()

    #Starting simulation at t != 0
    neuron.h.t = cell.tstart

    cell._loadspikes()

    #print sim.time at intervals
    cdef int counter = 0
    cdef double interval
    cdef double tstop = cell.tstop
    cdef double t0 = time()
    cdef double ti = neuron.h.t
    cdef double rtfactor
    if tstop >= 10000:
        interval = 1000. / cell.dt
    else:
        interval = 100. / cell.dt

    while neuron.h.t < tstop:
        neuron.h.fadvance()
        counter += 1
        if counter % interval == 0:
            rtfactor = (neuron.h.t - ti)  * 1E-3 / (time() - t0 + 1E-9)
            if cell.verbose:
                print('t = {:.0f}, realtime factor: {:.3f}'.format(neuron.h.t,
                                                                   rtfactor))
            t0 = time()
            ti = neuron.h.t


def _run_simulation_with_electrode(cell, cvode, electrode=None,
                                   variable_dt=False, atol=0.001,
                                   to_memory=True, to_file=False,
                                   file_name=None, dotprodcoeffs=None,
                                   rec_current_dipole_moment=False):
    """
    Running the actual simulation in NEURON.
    electrode argument used to determine coefficient
    matrix, and calculate the LFP on every time step.
    """

    #c-declare some variables
    cdef int i, j, tstep#, ncoeffs
    #cdef int totnsegs = cell.totnsegs
    cdef double tstop = cell.tstop
    cdef int counter
    cdef int lendotprodcoeffs0
    cdef double interval
    cdef double t0
    cdef double ti
    cdef double rtfactor
    cdef double dt = cell.dt
    cdef np.ndarray[DTYPE_t, ndim=2, negative_indices=False] coeffs
    cdef np.ndarray[DTYPE_t, ndim=2, negative_indices=False] current_dipole_moment
    cdef np.ndarray[DTYPE_t, ndim=2, negative_indices=False] midpoints

    #check if h5py exist and saving is possible
    try:
        import h5py
    except:
        print('h5py not found, LFP to file not possible')
        to_file = False
        file_name = None

    if variable_dt and to_file:
        print('LFP to file with variable_dt is not supported')
        to_file = False
        file_name = None

    # Use electrode object(s) to calculate coefficient matrices for LFP
    # calculations. If electrode is a list, then
    if cell.verbose:
        print('precalculating geometry - LFP mapping')

    #put electrodecoeff in a list, if it isn't already
    if dotprodcoeffs is not None:
        if type(dotprodcoeffs) != list:
            dotprodcoeffs = [dotprodcoeffs]
        electrodes = []
    else:
        #create empty list if no dotprodcoeffs are supplied
        dotprodcoeffs = []

    #just for safekeeping
    lendotprodcoeffs0 = len(dotprodcoeffs)

    #access electrode object and append mapping
    if electrode is not None:
        #put electrode argument in list if needed
        if type(electrode) == list:
            electrodes = electrode
        else:
            electrodes = [electrode]

        for el in electrodes:
            el.calc_mapping(cell)
            dotprodcoeffs.append(el.mapping)
    elif electrode is None:
        electrodes = None


    # Initialize NEURON simulations of cell object
    neuron.h.dt = cell.dt

    # needed for variable dt method
    if variable_dt:
        cvode.active(1)
        cvode.atol(atol)
    else:
        cvode.active(0)

    # re-initialize state
    neuron.h.finitialize(cell.v_init)

    # initialize current- and record
    if cvode.active():
        cvode.re_init()
    else:
        neuron.h.fcurrent()

    neuron.h.frecord_init()  # wrong voltages t=0 for tstart < 0 otherwise

    #Starting simulation at t != 0
    neuron.h.t = cell.tstart

    #load spike times from NetCon
    cell._loadspikes()

    #print sim.time at intervals
    counter = 0
    tstep = 0
    t0 = time()
    ti = neuron.h.t
    if tstop >= 10000:
        interval = 1000. / dt
    else:
        interval = 100. / dt

    #temp vector to store membrane currents at each timestep
    imem = np.zeros(cell.totnsegs)
    #LFPs for each electrode will be put here during simulation
    if to_memory:
        electrodesLFP = []
        for coeffs in dotprodcoeffs:
            if not cvode.active():
                # if cvode is on lists are created on-the-fly
                electrodesLFP.append(np.zeros((coeffs.shape[0],
                                     int(cell.tstop / cell.dt) + 1)))
    #LFPs for each electrode will be put here during simulations
    if to_file:
        #ensure right ending:
        file_name = Path(file_name)
        if file_name.suffix != '.h5':
            file_name = file_name.parent / (file_name.name + '.h5')
        if not cvode.active():
            el_LFP_file = h5py.File(file_name, 'w')
            i = 0
            for coeffs in dotprodcoeffs:
                el_LFP_file['electrode{:03d}'.format(i)] = np.zeros((coeffs.shape[0], int(cell.tstop / cell.dt + 1)))
                i += 1

    # create a 2D array representation of segment midpoints for dot product
    # with transmembrane currents when computing dipole moment
    if rec_current_dipole_moment:
        if not cvode.active():
            current_dipole_moment = cell.current_dipole_moment.copy()
            cell.current_dipole_moment = np.array([[]])
        midpoints = np.c_[cell.xmid, cell.ymid, cell.zmid]

    #run fadvance until time limit, and calculate LFPs for each timestep
    while neuron.h.t < tstop:
        if neuron.h.t >= 0:
            i = 0
            for sec in cell.allseclist:
                for seg in sec:
                    imem[i] = seg.i_membrane_
                    i += 1

            if rec_current_dipole_moment:
                if not cvode.active():
                    current_dipole_moment[tstep, ] = np.dot(imem, midpoints)
                else:
                    cell.current_dipole_moment.append(np.dot(imem, midpoints))

            if to_memory:
                for j, coeffs in enumerate(dotprodcoeffs):
                    if not cvode.active():
                        electrodesLFP[j][:, tstep] = np.dot(coeffs, imem)
                    else:
                        if tstep == 0:
                            electrodesLFP.append(np.dot(coeffs, imem)[:, np.newaxis])
                        else:
                            electrodesLFP[j] = np.hstack((electrodesLFP[j], np.dot(coeffs, imem)[:, np.newaxis]))
            if to_file:
                for j, coeffs in enumerate(dotprodcoeffs):
                    el_LFP_file['electrode{:03d}'.format(j)
                                ][:, tstep] = np.dot(coeffs, imem)

            tstep += 1
        neuron.h.fadvance()
        counter += 1
        if counter % interval == 0:
            rtfactor = (neuron.h.t - ti) * 1E-3 / (time() - t0)
            if cell.verbose:
                print('t = {:.0f}, realtime factor: {:.3f}'.format(neuron.h.t,
                                                                   rtfactor))
            t0 = time()
            ti = neuron.h.t

    if tstep < len(cell._neuron_tvec):
        #calculate LFP after final fadvance()
        i = 0
        for sec in cell.allseclist:
            for seg in sec:
                imem[i] = seg.i_membrane_
                i += 1

        if rec_current_dipole_moment:
            if not cvode.active():
                current_dipole_moment[tstep, ] = np.dot(imem, midpoints)
            else:
                cell.current_dipole_moment.append(np.dot(imem, midpoints))


        if to_memory:
            for j, coeffs in enumerate(dotprodcoeffs):
                if not cvode.active():
                    electrodesLFP[j][:, tstep] = np.dot(coeffs, imem)
                else:
                    electrodesLFP[j] = np.hstack((electrodesLFP[j], np.dot(coeffs, imem)[:, np.newaxis]))

        if to_file:
            for j, coeffs in enumerate(dotprodcoeffs):
                el_LFP_file['electrode{:03d}'.format(j)
                            ][:, tstep] = np.dot(coeffs, imem)

    # update current dipole moment values
    if rec_current_dipole_moment:
        if not cvode.active():
            cell.current_dipole_moment = np.array(current_dipole_moment)
        else:
            cell.current_dipole_moment = np.array(cell.current_dipole_moment)

    # Final step, put LFPs in the electrode object, superimpose if necessary
    # If electrode.perCellLFP, store individual LFPs
    if to_memory:
        #the first few belong to input dotprodcoeffs
        cell.dotprodresults = electrodesLFP[:lendotprodcoeffs0]
        #the remaining belong to input electrode arguments
        if electrodes is not None:
            for j, LFP in enumerate(electrodesLFP):
                if not j < lendotprodcoeffs0:
                    if hasattr(electrodes[j-lendotprodcoeffs0], 'LFP'):
                        electrodes[j-lendotprodcoeffs0].LFP += LFP
                    else:
                        electrodes[j-lendotprodcoeffs0].LFP = LFP
                    #will save each cell contribution separately
                    if electrodes[j-lendotprodcoeffs0].perCellLFP:
                        if not hasattr(electrodes[j], 'CellLFP'):
                            electrodes[j-lendotprodcoeffs0].CellLFP = []
                        electrodes[j-lendotprodcoeffs0].CellLFP.append(LFP)
                    electrodes[j-lendotprodcoeffs0].electrodecoeff = dotprodcoeffs[j]

    if to_file:
        el_LFP_file.close()


cpdef _collect_geometry_neuron(cell):
    """Loop over allseclist to determine area, diam, xyz-start- and
    endpoints, embed geometry to cell object"""


    cdef np.ndarray[DTYPE_t, ndim=1, negative_indices=False] areavec = np.zeros(cell.totnsegs)
    cdef np.ndarray[DTYPE_t, ndim=1, negative_indices=False] diamvec = np.zeros(cell.totnsegs)
    cdef np.ndarray[DTYPE_t, ndim=1, negative_indices=False] lengthvec = np.zeros(cell.totnsegs)

    cdef np.ndarray[DTYPE_t, ndim=1, negative_indices=False] xstartvec = np.zeros(cell.totnsegs)
    cdef np.ndarray[DTYPE_t, ndim=1, negative_indices=False] xendvec = np.zeros(cell.totnsegs)
    cdef np.ndarray[DTYPE_t, ndim=1, negative_indices=False] ystartvec = np.zeros(cell.totnsegs)
    cdef np.ndarray[DTYPE_t, ndim=1, negative_indices=False] yendvec = np.zeros(cell.totnsegs)
    cdef np.ndarray[DTYPE_t, ndim=1, negative_indices=False] zstartvec = np.zeros(cell.totnsegs)
    cdef np.ndarray[DTYPE_t, ndim=1, negative_indices=False] zendvec = np.zeros(cell.totnsegs)

    cdef DTYPE_t gsen2, secL
    cdef LTYPE_t counter, nseg, n3d, i


    cdef np.ndarray[DTYPE_t, ndim=1, negative_indices=False] L, x, y, z, segx, segx0, segx1


    counter = 0

    #loop over all segments
    for sec in cell.allseclist:
        n3d = int(neuron.h.n3d())
        nseg = sec.nseg
        gsen2 = 1./2/nseg
        secL = sec.L
        if n3d > 0:
            #create interpolation objects for the xyz pt3d info:
            L = np.zeros(n3d)
            x = np.zeros(n3d)
            y = np.zeros(n3d)
            z = np.zeros(n3d)
            for i in range(n3d):
                L[i] = neuron.h.arc3d(i)
                x[i] = neuron.h.x3d(i)
                y[i] = neuron.h.y3d(i)
                z[i] = neuron.h.z3d(i)

            #normalize as seg.x [0, 1]
            L /= secL

            #temporary store position of segment midpoints
            segx = np.zeros(nseg)
            i = 0
            for seg in sec:
                segx[i] = seg.x
                i += 1

            #can't be >0 which may happen due to NEURON->Python float transfer:
            #segx0 = (segx - gsen2).round(decimals=6)
            #segx1 = (segx + gsen2).round(decimals=6)
            segx0 = segx - gsen2
            segx1 = segx + gsen2

            #fill vectors with interpolated coordinates of start and end points
            xstartvec[counter:counter+nseg] = np.interp(segx0, L, x)
            xendvec[counter:counter+nseg] = np.interp(segx1, L, x)

            ystartvec[counter:counter+nseg] = np.interp(segx0, L, y)
            yendvec[counter:counter+nseg] = np.interp(segx1, L, y)

            zstartvec[counter:counter+nseg] = np.interp(segx0, L, z)
            zendvec[counter:counter+nseg] = np.interp(segx1, L, z)

            #fill in values area, diam, length
            for seg in sec:
                areavec[counter] = neuron.h.area(seg.x)
                diamvec[counter] = seg.diam
                lengthvec[counter] = secL/nseg

                counter += 1


    #set cell attributes
    cell.xstart = xstartvec
    cell.ystart = ystartvec
    cell.zstart = zstartvec

    cell.xend = xendvec
    cell.yend = yendvec
    cell.zend = zendvec

    cell.area = areavec
    cell.diam = diamvec
    cell.length = lengthvec



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Copyright (C) 2012 Computational Neuroscience Group, NMBU.

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

"""


import numpy as np
import scipy.signal as ss

def load(filename):
    """Generic loading of cPickled objects from file"""
    import pickle
    
    filen = open(filename,'rb')
    obj = pickle.load(filen)
    filen.close()
    return obj

def noise_brown(ncols, nrows=1, weight=1., filter=None, filterargs=None):
    """Return 1/f^2 noise of shape(nrows, ncols obtained by taking
    the cumulative sum of gaussian white noise, with rms weight.
    
    If filter is not None, this function will apply
    the filter coefficients obtained
    by:
        
    >>> b, a = filter(**filterargs)
    >>> signal = scipy.signal.lfilter(b, a, signal)
    """
    def rms_flat(a):
        """
        Return the root mean square of all the elements of *a*, flattened out.
        """
        return np.sqrt(np.mean(np.absolute(a)**2))

    if filter is not None:
        coeff_b, coeff_a = list(filter(**filterargs))
    
    noise = np.zeros((nrows, ncols))    
    for i in range(nrows):
        signal = np.random.normal(size=ncols+10000).cumsum()
        if filter is not None:
            signal = ss.lfilter(coeff_b, coeff_a, signal)
        noise[i, :] = signal[10000:]
        noise[i, :] /= rms_flat(noise[i, :])
        noise[i, :] *= weight
    return noise




#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""LFPy setuptools file

"""

import os
import sys
import shutil

d = {}
exec(open("LFPy/version.py").read(), None, d)
version = d['version']

try:
    from setuptools import setup, Extension
except ImportError as ie:
    raise ie('please install setuptools')
try:
    import numpy
    from Cython.Distutils import build_ext
    cmdclass = {'build_ext': build_ext}
    ext_modules = [Extension('LFPy.run_simulation',
                   [os.path.join('LFPy', 'run_simulation.pyx')],
                   include_dirs=[numpy.get_include()]),
                   Extension('LFPy.alias_method',
                   [os.path.join('LFPy', 'alias_method.pyx')],
                   include_dirs=[numpy.get_include()]),
                   ]
except ImportError:
    print("'from Cython.Distutils import build_ext' or 'import numpy' failed!")
    print("Cython extensions will not be compiled, and")
    print("simulations in LFPy may run slower")
    cmdclass = {}
    ext_modules = []

# try and locate the nrnivmodl script of NEURON in PATH so that the
# NEURON extension file LFPy/sinsyn.mod can be compiled in place and be copied
# as part of the package_data, allowing unit tests to run
from distutils.spawn import find_executable, spawn
if not any(arg in sys.argv for arg in ['sdist', 'upload']):
    if find_executable('nrnivmodl') is not None:
        os.chdir(os.path.join('LFPy', 'test'))
        for path in ['x86_64']:
            if os.path.isdir(path):
                shutil.rmtree(path)
        spawn([find_executable('nrnivmodl')])
        os.chdir(os.path.join('..', '..'))
    else:
        print("nrnivmodl script not found in PATH, thus NEURON .mod files" +
              "could not be compiled, and LFPy.test() functions will fail")


with open('README.md') as file:
    long_description = file.read()

setup(
    name="LFPy",
    version=version,
    maintainer="LFPy-team",
    maintainer_email='lfpy@users.noreply.github.com',
    packages=['LFPy'],
    package_data={'LFPy': ['*.pyx',
                           os.path.join('test', '*.hoc'),
                           os.path.join('test', '*.py'),
                           os.path.join('test', 'sinsyn.mod'),
                           ]},
    include_package_data=True,
    cmdclass=cmdclass,
    ext_modules=ext_modules,
    url='http://LFPy.readthedocs.io',
    download_url='https://github.com/LFPy/LFPy/tarball/v2.0.7',
    license='LICENSE',
    description='A module for modeling extracellular potentials of multicompartment neuron models built on NEURON',
    long_description=long_description,
    classifiers=[
        'License :: OSI Approved :: GNU General Public License (GPL)',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Cython',
        'Operating System :: OS Independent',
        'Topic :: Scientific/Engineering',
        'Topic :: Scientific/Engineering :: Physics',
        'Topic :: Utilities',
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'Development Status :: 5 - Production/Stable',
        ],
    install_requires=[
        'numpy>=1.8',
        'scipy>=0.14',
        'Cython>=0.20',
        'h5py>=2.5',
        'mpi4py>=1.2',
        'MEAutility>=1.4.2'
        ],
    extras_require={'tests': ['pytest'],
                    'docs': ['sphinx', 'numpydoc', 'sphinx_rtd_theme']
                    },
    dependency_links=[],
    provides=['LFPy'],
    )



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Copyright (C) 2012 Computational Neuroscience Group, NMBU.

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

"""
import os
import numpy as np
from scipy.integrate import quad
from numpy import real, imag
import LFPy
import neuron


######## Functions used by tests: ##############################################
def stickSimulation(method):
    stickParams = {
        'morphology' : os.path.join(LFPy.__path__[0], 'test', 'stick.hoc'),
        'cm' : 1,
        'Ra' : 150,
        'v_init' : -65,
        'passive' : True,
        'passive_parameters' : {'g_pas' : 1./30000, 'e_pas' : -65},
        'tstart' : -100,
        'tstop' : 100,
        'dt' : 2**-6,
        'nsegs_method' : 'lambda_f',
        'lambda_f' : 1000,

    }

    electrodeParams = {
        'sigma' : 0.3,
        'x' : np.ones(11) * 100.,
        'y' : np.zeros(11),
        'z' : np.linspace(1000, 0, 11),
        'method' : method
    }

    stimParams = {
        'pptype' : 'SinSyn',
        'delay' : -100.,
        'dur' : 1000.,
        'pkamp' : 1.,
        'freq' : 100.,
        'phase' : -np.pi/2,
        'bias' : 0.,
        'record_current' : True
    }


    electrode = LFPy.RecExtElectrode(**electrodeParams)

    stick = LFPy.Cell(**stickParams)
    stick.set_pos(z=-stick.zstart[0])

    synapse = LFPy.StimIntElectrode(stick, stick.get_closest_idx(0, 0, 1000),
                           **stimParams)
    stick.simulate(electrode, rec_imem=True, rec_vmem=True)

    return electrode.LFP

def stickSimulationAveragingElectrode(contactRadius, contactNPoints, method):
    stickParams = {
        'morphology' : os.path.join(LFPy.__path__[0], 'test', 'stick.hoc'),
        'cm' : 1,
        'Ra' : 150,
        'v_init' : -65,
        'passive' : True,
        'passive_parameters' : {'g_pas' : 1./30000, 'e_pas' : -65},
        'tstart' : -100,
        'tstop' : 100,
        'dt' : 2**-6,
        'nsegs_method' : 'lambda_f',
        'lambda_f' : 1000,

    }

    N = np.empty((11, 3))
    for i in range(N.shape[0]): N[i,] = [1, 0, 0] #normal unit vec. to contacts
    electrodeParams = {
        'sigma' : 0.3,
        'x' : np.ones(11) * 100.,
        'y' : np.zeros(11),
        'z' : np.linspace(1000, 0, 11),
        'r' : contactRadius,
        'n' : 10,
        'N' : N,
        'method' : method
    }

    stimParams = {
        'pptype' : 'SinSyn',
        'delay' : -100.,
        'dur' : 1000.,
        'pkamp' : 1.,
        'freq' : 100.,
        'phase' : -np.pi/2,
        'bias' : 0.,
        'record_current' : True
    }


    electrode = LFPy.RecExtElectrode(**electrodeParams)

    stick = LFPy.Cell(**stickParams)
    stick.set_pos(z=-stick.zstart[0])

    synapse = LFPy.StimIntElectrode(stick, stick.get_closest_idx(0, 0, 1000),
                           **stimParams)
    stick.simulate(electrode, rec_imem=True, rec_vmem=True)

    return electrode.LFP

def stickSimulationDotprodcoeffs(method):
    stickParams = {
        'morphology' : os.path.join(LFPy.__path__[0], 'test', 'stick.hoc'),
        'cm' : 1,
        'Ra' : 150,
        'v_init' : -65,
        'passive' : True,
        'passive_parameters' : {'g_pas' : 1./30000, 'e_pas' : -65},
        'tstart' : -100,
        'tstop' : 100,
        'dt' : 2**-6,
        'nsegs_method' : 'lambda_f',
        'lambda_f' : 1000,

    }

    electrodeParams = {
        'sigma' : 0.3,
        'x' : np.ones(11) * 100.,
        'y' : np.zeros(11),
        'z' : np.linspace(1000, 0, 11),
        'method' : method
    }

    stimParams = {
        'pptype' : 'SinSyn',
        'delay' : -100.,
        'dur' : 1000.,
        'pkamp' : 1.,
        'freq' : 100.,
        'phase' : -np.pi/2,
        'bias' : 0.,
        'record_current' : True
    }



    stick = LFPy.Cell(**stickParams)
    stick.set_pos(z=-stick.zstart[0])

    #dummy variables for mapping
    stick.imem = np.eye(stick.totnsegs)
    stick.tvec = np.arange(stick.totnsegs)*stick.dt

    electrode = LFPy.RecExtElectrode(stick, **electrodeParams)
    electrode.calc_lfp()
    #not needed anymore:
    del stick.imem, stick.tvec

    synapse = LFPy.StimIntElectrode(stick, stick.get_closest_idx(0, 0, 1000),
                           **stimParams)
    stick.simulate(dotprodcoeffs=electrode.LFP,
                   rec_imem=True, rec_vmem=True)

    return stick.dotprodresults[0]


def analytical_LFP(time=np.linspace(0, 100, 1001),
                   stickLength=1000.,
                   stickDiam=2.,
                   Rm=30000.,
                   Cm=1.,
                   Ri=150.,
                   stimFrequency=100.,
                   stimAmplitude=1.,
                   # stimPos=1.,
                   sigma=0.3,
                   electrodeR=100.,
                   electrodeZ=0.,
                   ):
    """
    Will calculate the analytical LFP from a dendrite stick aligned with z-axis.
    The synaptic current is always assumed to be at the end of the stick, i.e.
    Zin = stickLength.

    Parameters
    ----------
    time : ndarray
        The LFP is calculated for values in this np.array (ms)
    stickLength : float
        length of stick (mum)
    stickDiam : float
        diameter of stick (mum)
    Rm : float
        Membrane resistivity (Ohm * cm2)
    Cm : float
        Membrane capacitance (muF/cm2)
    Ri : float
        Intracellular resistivity (Ohm*cm)
    stimFrequency : float
        Frequency of cosine synapse current (Hz)
    stimAmplitude : float
        Amplitude of cosine synapse current (nA)
    # stimPos : float in [0, 1]
    #     Relative stimulus current position from start (0) to end (1) of stick
    sigma : float
        Extracellular conductivity (muS/mum)
    electrodeR : float
        Radial distance from stick (mum)
    electrodeZ : float
        Longitudal distance along stick(mum)
    """
    Gm = 1. / Rm            # specific membrane conductivity (S/cm2)
    gm = 1E2 * np.pi * stickDiam / Rm     # absolute membrane conductance (muS / mum)
    ri = 1E-2 * 4. * Ri / (np.pi * stickDiam**2) # intracellular resistance  (Mohm/mum)

    Lambda = 1E2 / np.sqrt(gm * ri) # Electrotonic length constant of stick (mum)
    Ginf = 10 / (ri * Lambda)   # infinite stick input cond (10*muS)?

    tau_m = Rm * Cm / 1000        # membrane time constant (ms)
    Omega = 2 * np.pi * stimFrequency * tau_m / 1000 #impedance
    Zel = electrodeZ / Lambda    # z-position of extracellular point, in units of Lambda
    L = stickLength / Lambda      # Length of stick in units of Lambda
    Rel = electrodeR / Lambda    # extracellular, location along x-axis, or radius, in units of Lambda
    q = np.sqrt(1 + 1j*Omega)	    # Note: j is sqrt(-1)
    Yin = q * Ginf * np.tanh(q * L)	    # Admittance
    Zin = stickLength / Lambda  # unitless location of input current
    # Zin = stickLength / Lambda * stimPos  # unitless location of input current

    PhiExImem = np.empty(time.size)
    PhiExInput = np.empty(time.size)

    def i_mem(z): #z is location at stick
        return gm * q**2 * np.cosh(q * z) / np.cosh(q * L) * stimAmplitude / Yin

    def f_to_integrate(z):
        return 1E-3 / (4 * np.pi * sigma) * i_mem(z) \
            / np.sqrt(Rel**2 + (z - Zel)**2)

    #calculate contrib from membrane currents
    Vex_imem = -complex_quadrature(f_to_integrate, 0, L, epsabs=1E-20)

    #adding contrib from input current to Vex
    Vex_input = stimAmplitude / (4 * np.pi * sigma * Lambda * np.sqrt(Rel**2 + (Zin-Zel)**2))

    PhiExImemComplex = Vex_imem * np.exp(1j * 2 * np.pi * stimFrequency *
                                              time / 1000)
    PhiExInputComplex = Vex_input * np.exp(1j * 2 * np.pi * stimFrequency *
                                             time / 1000)

    #Using only real component
    PhiExImem = PhiExImemComplex.real
    PhiExInput = PhiExInputComplex.real

    PhiEx = PhiExImem + PhiExInput
    return PhiEx

def complex_quadrature(func, a, b, **kwargs):
    """
    Will return the complex integral value.
    """
    def real_func(x):
        return real(func(x))
    def imag_func(x):
        return imag(func(x))
    real_integral = quad(real_func, a, b, **kwargs)
    imag_integral = quad(imag_func, a, b, **kwargs)
    return real_integral[0] + 1j*imag_integral[0]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Example script loading and execiting a simulation with the
Hay et al. 2011 L5b-pyramidal cell model, which is implemented by default
using templates.

This script assume that the model files is downloaded and unzipped inside
this folder from ModelDB:
http://senselab.med.yale.edu/modeldb/ShowModel.asp?model=139653

The mod-files inside /L5bPCmodelsEH/mod/ must be compiled using nrnivmodl.

Note that LFPy can only deal with one cell at the time, creating several
cell objects will slow everything down, but each cell *should* get the correct
cell responses.

Execution:

    python example_loadL5bPCmodelsEH.py

Copyright (C) 2017 Computational Neuroscience Group, NMBU.

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
"""

import LFPy
import neuron
import matplotlib.pyplot as plt
import os
import sys
if sys.version < '3':
    from urllib2 import urlopen
else:    
    from urllib.request import urlopen
import zipfile
import ssl
from warnings import warn



#Fetch Hay et al. 2011 model files
if not os.path.isfile('L5bPCmodelsEH/morphologies/cell1.asc'):
    #get the model files:
    u = urlopen('http://senselab.med.yale.edu/ModelDB/eavBinDown.asp?o=139653&a=23&mime=application/zip',
                context=ssl._create_unverified_context())
    localFile = open('L5bPCmodelsEH.zip', 'w')
    localFile.write(u.read())
    localFile.close()
    #unzip:
    myzip = zipfile.ZipFile('L5bPCmodelsEH.zip', 'r')
    myzip.extractall('.')
    myzip.close()

#compile mod files every time, because of incompatibility with Mainen96 files:
if "win32" in sys.platform:
    pth = "L5bPCmodelsEH/mod/"
    warn("no autompile of NMODL (.mod) files on Windows.\n" 
         + "Run mknrndll from NEURON bash in the folder L5bPCmodelsEH/mod and rerun example script")
    if not pth in neuron.nrn_dll_loaded:
        neuron.h.nrn_load_dll(pth+"nrnmech.dll")
    neuron.nrn_dll_loaded.append(pth)
else:
    os.system('''
              cd L5bPCmodelsEH/mod/
              nrnivmodl
              ''')
    neuron.load_mechanisms('L5bPCmodelsEH/mod/')

#remove cells from previous script executions
neuron.h('forall delete_section()')

#cell parameters with additional arguments for the TemplateCell-class.
#Note that 'morphology' is required, even though it is loaded through
#'templateargs'!
#Reason is LFPy looks for a default rotation .rot-file.
cellParams = {
    'morphology'    : 'L5bPCmodelsEH/morphologies/cell1.asc',
    'templatefile'  : ['L5bPCmodelsEH/models/L5PCbiophys3.hoc',
                       'L5bPCmodelsEH/models/L5PCtemplate.hoc'],
    'templatename'  : 'L5PCtemplate',
    'templateargs'  : 'L5bPCmodelsEH/morphologies/cell1.asc',
    'nsegs_method' : None,
    'v_init' : -80,
    'tstart' : 0,
    'tstop' : 3000,
    'dt' : 2**-3,
    'verbose' : True,
    'extracellular' : False,
}

#Use the TemplateCell-class to create the cell
cell = LFPy.TemplateCell(**cellParams)

#some stimuli
PointProcParams = {
    'idx' : 0,
    'record_current' : False,
    'pptype' : 'IClamp',
    'amp' : 0.793,
    'dur' : 2000,
    'delay' : 700,
}

pointProcess = LFPy.StimIntElectrode(cell, **PointProcParams)

#run simulation
cell.simulate(rec_variables = [])

#plot response
plt.plot(cell.tvec, cell.somav)
plt.show()


#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Copyright (C) 2012 Computational Neuroscience Group, NMBU.

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

"""


import os
import unittest
import numpy as np
import LFPy


class testAliasMethod(unittest.TestCase):
    """
    test LFPy.alias_method methods
    """


    def test_alias_method_00(self):
        """deterministic probabilities 0.0 and 1.0"""
        idx = np.arange(2)
        probs = np.arange(2).astype(float)
        nidx = 1000000
        bins = np.arange(3)

        hist, _ = np.histogram(LFPy.alias_method.alias_method(idx, probs, nidx),
                               bins)

        self.assertEqual(nidx, hist[1])

    def test_alias_method_01(self):
        """probabilities 0.25 and 0.75"""
        idx = np.arange(2)
        probs = np.array([0.25, 0.75])
        nidx = 1000000
        bins = np.arange(3)

        hist, _ = np.histogram(LFPy.alias_method.alias_method(idx, probs, nidx),
                               bins)

        # compute Pearson correlation coefficients between area and histogram
        # reporting success if within 7 decimal places
        self.assertAlmostEqual(np.corrcoef(probs, hist.astype(float))[0, 1], 1., places=7)

    def test_alias_method_02(self):
        """probabilities 0.75 and 0.25"""
        idx = np.arange(2)
        probs = np.array([0.75, 0.25])
        nidx = 1000000
        bins = np.arange(3)

        hist, _ = np.histogram(LFPy.alias_method.alias_method(idx, probs, nidx),
                               bins)

        # compute Pearson correlation coefficients between area and histogram
        # reporting success if within 7 decimal places
        self.assertAlmostEqual(np.corrcoef(probs, hist.astype(float))[0, 1], 1., places=7)

    def test_alias_method_03(self):
        '''over range of normalized probabilities'''
        size = 5
        idx = np.arange(size)
        probs = np.arange(size).astype(float)**2
        probs /= probs.sum()
        nidx = 1000000
        bins = np.arange(probs.size + 1)

        hist, _ = np.histogram(LFPy.alias_method.alias_method(idx, probs, nidx),
                               bins)

        # compute Pearson correlation coefficients between area and histogram
        # reporting success if within 5 decimal places
        self.assertAlmostEqual(np.corrcoef(probs, hist.astype(float))[0, 1], 1., places=4)


    def test_alias_method_04(self):
        """deterministic probabilities 1.0 and 0.0"""
        idx = np.arange(2)
        probs = np.arange(2).astype(float)[::-1]
        nidx = 1000000
        bins = np.arange(3)

        hist, _ = np.histogram(LFPy.alias_method.alias_method(idx, probs, nidx),
                               bins)

        self.assertEqual(nidx, hist[0])



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Copyright (C) 2012 Computational Neuroscience Group, NMBU.

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

"""

import sys
import os
import posixpath
import unittest
import numpy as np
import LFPy
import neuron

# for nosetests to run load mechanisms
if "win32" in sys.platform:
    pth = os.path.join(LFPy.__path__[0], 'test', 'nrnmech.dll')
    pth = pth.replace(os.sep, posixpath.sep)
    if not pth in neuron.nrn_dll_loaded:
        neuron.h.nrn_load_dll(pth)
        neuron.nrn_dll_loaded.append(pth)
else:
    neuron.load_mechanisms(os.path.join(LFPy.__path__[0], 'test'))

class testPointProcess(unittest.TestCase):
    """
    test class LFPy.PointProcess
    """
    def test_PointProcess_00(self):
        cell = LFPy.Cell(morphology=os.path.join(LFPy.__path__[0], 'test',
                                                 'ball_and_sticks.hoc'))
        pp = LFPy.PointProcess(cell=cell, idx=0)
        self.assertTrue(np.alltrue(np.array([pp.x, pp.y, pp.z])==cell.somapos))
        self.assertEqual(pp.idx, 0)
    

class testSynapse(unittest.TestCase):
    """
    test class LFPy.Synapse
    """
    def test_Synapse_00(self):
        cell = LFPy.Cell(morphology=os.path.join(LFPy.__path__[0], 'test',
                                                 'ball_and_sticks.hoc'))
        syn = LFPy.Synapse(cell=cell, idx=0, syntype='ExpSynI',
                           weight=1., tau=5., record_current=True,
                           record_potential=True)
        syn.set_spike_times(np.array([10.]))
        cell.simulate()

        i = np.zeros(cell.tvec.size)
        i[cell.tvec > 10.] = -np.exp(-np.arange((cell.tvec > 10.).sum())*cell.dt / 5.)

        np.testing.assert_allclose(i, syn.i, rtol=1E-1)
        np.testing.assert_equal(cell.somav, syn.v)
        

    def test_Synapse_01(self):
        cell = LFPy.Cell(morphology=os.path.join(LFPy.__path__[0], 'test',
                                                 'ball_and_sticks.hoc'))
        syn0 = LFPy.Synapse(cell=cell, idx=0, syntype='ExpSynI',
                           weight=1., tau=5., record_current=True,
                           record_potential=True)
        syn0.set_spike_times(np.array([10.]))

        syn1 = LFPy.Synapse(cell=cell, idx=1, syntype='ExpSynI',
                           weight=1., tau=5., record_current=True,
                           record_potential=False)
        syn1.set_spike_times(np.array([20.]))

        syn2 = LFPy.Synapse(cell=cell, idx=2, syntype='ExpSynI',
                           weight=1., tau=5., record_current=False,
                           record_potential=True)
        syn2.set_spike_times(np.array([30.]))

        syn3 = LFPy.Synapse(cell=cell, idx=3, syntype='ExpSynI',
                           weight=1., tau=5., record_current=False,
                           record_potential=False)
        syn3.set_spike_times(np.array([40.]))

        cell.simulate()
        
        i = np.zeros(cell.tvec.size)
        i[cell.tvec > 10.] = -np.exp(-np.arange((cell.tvec > 10.).sum())*cell.dt / 5.)

        np.testing.assert_allclose(i, syn0.i, rtol=1E-1)
        np.testing.assert_equal(cell.somav, syn0.v)

        self.assertTrue(hasattr(syn1, 'i'))
        i = np.zeros(cell.tvec.size)
        i[cell.tvec > 20.] = -np.exp(-np.arange((cell.tvec > 20.).sum())*cell.dt / 5.)
        self.assertFalse(hasattr(syn1, 'v'))
        np.testing.assert_allclose(i, syn1.i, rtol=1E-1)
        
        self.assertFalse(hasattr(syn2, 'i'))
        self.assertTrue(hasattr(syn2, 'v'))
        
        self.assertFalse(hasattr(syn3, 'i'))
        self.assertFalse(hasattr(syn3, 'v'))


    def test_Synapse_02(self):
        cell = LFPy.Cell(morphology=os.path.join(LFPy.__path__[0], 'test',
                                                 'ball_and_sticks.hoc'))
        t0 = 10.
        t1 = 30.
        tau = 5.
        syn0 = LFPy.Synapse(cell=cell, idx=0, syntype='ExpSynI',
                           weight=1., tau=tau, record_current=True,
                           record_potential=True)
        syn0.set_spike_times_w_netstim(noise=0., start=t0-1, number=1) # -1 to acct for delay

        syn1 = LFPy.Synapse(cell=cell, idx=0, syntype='ExpSynI',
                           weight=1., tau=tau, record_current=True,
                           record_potential=True)
        syn1.set_spike_times(np.array([t1]))

        cell.simulate()
        
        i0 = np.zeros(cell.tvec.size)
        i0[cell.tvec > t0] = -np.exp(-np.arange((cell.tvec > t0).sum())*cell.dt / tau)

        i1 = np.zeros(cell.tvec.size)
        i1[cell.tvec > t1] = -np.exp(-np.arange((cell.tvec > t1).sum())*cell.dt / tau)

        np.testing.assert_allclose(i0, syn0.i, rtol=1E-1)
        np.testing.assert_equal(cell.somav, syn0.v)

        np.testing.assert_allclose(i1, syn1.i, rtol=1E-1)
        np.testing.assert_equal(cell.somav, syn1.v)


    def test_Synapse_03(self):
        cell = LFPy.Cell(morphology=os.path.join(LFPy.__path__[0], 'test',
                                                 'ball_and_sticks.hoc'))
        t0 = 10.
        t1 = 30.
        tau = 5.
        syn0 = LFPy.Synapse(cell=cell, idx=0, syntype='ExpSynI',
                           weight=1., tau=tau, record_current=True,
                           record_potential=True)
        syn0.set_spike_times_w_netstim(noise=0., start=t0-1, number=1) # -1 to acct for delay

        syn1 = LFPy.Synapse(cell=cell, idx=0, syntype='ExpSynI',
                           weight=1., tau=tau, record_current=True,
                           record_potential=True)
        syn1.set_spike_times_w_netstim(noise=0., start=t1-1, number=1) # -1 to acct for delay

        cell.simulate()
        
        i0 = np.zeros(cell.tvec.size)
        i0[cell.tvec > t0] = -np.exp(-np.arange((cell.tvec > t0).sum())*cell.dt / tau)

        i1 = np.zeros(cell.tvec.size)
        i1[cell.tvec > t1] = -np.exp(-np.arange((cell.tvec > t1).sum())*cell.dt / tau)

        np.testing.assert_allclose(i0, syn0.i, rtol=1E-1)
        np.testing.assert_equal(cell.somav, syn0.v)

        np.testing.assert_allclose(i1, syn1.i, rtol=1E-1)
        np.testing.assert_equal(cell.somav, syn1.v)


    def test_Synapse_04(self):
        cell = LFPy.Cell(morphology=os.path.join(LFPy.__path__[0], 'test',
                                                 'ball_and_sticks.hoc'))
        t0 = 10.
        t1 = 30.
        tau = 5.
        syn0 = LFPy.Synapse(cell=cell, idx=0, syntype='ExpSynI',
                           weight=1., tau=tau, record_current=True,
                           record_potential=True)

        syn1 = LFPy.Synapse(cell=cell, idx=0, syntype='ExpSynI',
                           weight=1., tau=tau, record_current=True,
                           record_potential=True)

        syn0.set_spike_times_w_netstim(noise=0., start=t0-1, number=1) # -1 to acct for delay
        syn1.set_spike_times(np.array([t1]))

        cell.simulate()
        
        i0 = np.zeros(cell.tvec.size)
        i0[cell.tvec > t0] = -np.exp(-np.arange((cell.tvec > t0).sum())*cell.dt / tau)

        i1 = np.zeros(cell.tvec.size)
        i1[cell.tvec > t1] = -np.exp(-np.arange((cell.tvec > t1).sum())*cell.dt / tau)

        np.testing.assert_allclose(i0, syn0.i, rtol=1E-1)
        np.testing.assert_equal(cell.somav, syn0.v)

        np.testing.assert_allclose(i1, syn1.i, rtol=1E-1)
        np.testing.assert_equal(cell.somav, syn1.v)
        
        
    def test_Synapse_05(self):
        cell = LFPy.Cell(morphology=os.path.join(LFPy.__path__[0], 'test',
                                                 'ball_and_sticks.hoc'))
        t0 = 10.
        t1 = 30.
        tau = 5.
        syn0 = LFPy.Synapse(cell=cell, idx=0, syntype='ExpSynI',
                           weight=1., tau=tau, record_current=True,
                           record_potential=True)

        syn1 = LFPy.Synapse(cell=cell, idx=0, syntype='ExpSynI',
                           weight=1., tau=tau, record_current=True,
                           record_potential=True)

        syn1.set_spike_times(np.array([t1]))
        syn0.set_spike_times_w_netstim(noise=0., start=t0-1, number=1) # -1 to acct for delay

        cell.simulate()
        
        i0 = np.zeros(cell.tvec.size)
        i0[cell.tvec > t0] = -np.exp(-np.arange((cell.tvec > t0).sum())*cell.dt / tau)

        i1 = np.zeros(cell.tvec.size)
        i1[cell.tvec > t1] = -np.exp(-np.arange((cell.tvec > t1).sum())*cell.dt / tau)

        np.testing.assert_allclose(i0, syn0.i, rtol=1E-1)
        np.testing.assert_equal(cell.somav, syn0.v)

        np.testing.assert_allclose(i1, syn1.i, rtol=1E-1)
        np.testing.assert_equal(cell.somav, syn1.v)


class testStimIntElectrode(unittest.TestCase):
    """
    test class LFPy.StimIntElectrode
    """
    def test_StimIntElectrode_00(self):
        cell = LFPy.Cell(morphology=os.path.join(LFPy.__path__[0], 'test',
                                                 'ball_and_sticks.hoc'))
        stim = LFPy.StimIntElectrode(cell=cell, idx=0, pptype='IClamp',
                            amp=1., dur=20., delay=10.,
                            record_potential=True,
                            record_current=True)
        cell.simulate()
        gt = np.zeros(cell.tvec.size)
        gt[(cell.tvec > 10.) & (cell.tvec <= 30.)] = 1.
        np.testing.assert_equal(gt, stim.i)
        np.testing.assert_equal(cell.somav, stim.v)

    def test_StimIntElectrode_01(self):
        cell = LFPy.Cell(morphology=os.path.join(LFPy.__path__[0], 'test',
                                                 'ball_and_sticks.hoc'), dt=1.,
                         v_init=-65.,
                         Ra = 150.,
                         cm = 1.,
                         passive=True,
                         passive_parameters=dict(g_pas=1./30000, e_pas=-65)
                         )
        stim = LFPy.StimIntElectrode(cell=cell,
                                     record_potential=True,
                                     **{'idx' : 0,
                                        'pptype' : 'VClamp',
                                        'amp[0]' : -65,
                                        'dur[0]' : 10,
                                        'amp[1]' : -55.,
                                        'dur[1]' : 20,
                                        'amp[2]' : -65,
                                        'dur[2]' : 10,
                                   })
        cell.simulate()
        gt = np.zeros(cell.tvec.size)-65.
        gt[(cell.tvec > 10.) & (cell.tvec <= 30.)] = -55.
        np.testing.assert_allclose(gt, cell.somav, rtol=1E-3)
        np.testing.assert_equal(cell.somav, stim.v)

    def test_StimIntElectrode_02(self):
        cell = LFPy.Cell(morphology=os.path.join(LFPy.__path__[0], 'test',
                                                 'ball_and_sticks.hoc'), dt=1.,
                         v_init=-65.,
                         Ra = 150.,
                         cm = 1.,
                         passive=True,
                         passive_parameters=dict(g_pas=1./30000, e_pas=-65)
                         )
        stim = LFPy.StimIntElectrode(cell=cell,
                                     record_potential=True,
                                     **{'idx' : 0,
                                        'pptype' : 'SEClamp',
                                        'amp1' : -65,
                                        'dur1' : 10,
                                        'amp2' : -55.,
                                        'dur2' : 20,
                                        'amp3' : -65,
                                        'dur3' : 10,
                                   })
        cell.simulate()
        gt = np.zeros(cell.tvec.size)-65.
        gt[(cell.tvec > 10.) & (cell.tvec <= 30.)] = -55.
        np.testing.assert_allclose(gt, cell.somav, rtol=1E-2)
        np.testing.assert_equal(cell.somav, stim.v)


    def test_StimIntElectrode_03(self):
        cell = LFPy.Cell(morphology=os.path.join(LFPy.__path__[0], 'test',
                                                 'ball_and_sticks.hoc'))
        stim0 = LFPy.StimIntElectrode(cell=cell, idx=0, pptype='IClamp',
                            amp=1., dur=20., delay=10.,
                            record_potential=True,
                            record_current=True)

        stim1 = LFPy.StimIntElectrode(cell=cell, idx=1, pptype='IClamp',
                            amp=1., dur=20., delay=30.,
                            record_potential=True,
                            record_current=False)

        stim2 = LFPy.StimIntElectrode(cell=cell, idx=2, pptype='IClamp',
                            amp=1., dur=20., delay=50.,
                            record_potential=False,
                            record_current=True)

        stim3 = LFPy.StimIntElectrode(cell=cell, idx=3, pptype='IClamp',
                            amp=1., dur=20., delay=70.,
                            record_potential=False,
                            record_current=False)


        cell.simulate()
        gt = np.zeros(cell.tvec.size)
        gt[(cell.tvec > 10.) & (cell.tvec <= 30.)] = 1.
        np.testing.assert_equal(gt, stim0.i)
        np.testing.assert_equal(cell.somav, stim0.v)
        
        self.assertTrue(hasattr(stim1, 'v'))
        self.assertTrue(cell.tvec.shape == stim1.v.shape)
        self.assertFalse(hasattr(stim2, 'v'))
        self.assertFalse(hasattr(stim3, 'v'))
        self.assertFalse(hasattr(stim1, 'i'))
        self.assertTrue(hasattr(stim2, 'i'))
        self.assertTrue(cell.tvec.shape == stim2.i.shape)        
        self.assertFalse(hasattr(stim3, 'i'))
        
  
    



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Copyright (C) 2012 Computational Neuroscience Group, NMBU.

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

"""


import os
import unittest
import numpy as np
import LFPy
import pickle

class testTools(unittest.TestCase):
    """
    test LFPy.tools methods
    """
    def test_tools_load_00(self):
        filename = 'test.cpickle'
        x = object()
        f = open(filename, 'wb')
        pickle.dump(x, f)
        f.close()
        self.assertTrue(isinstance(LFPy.tools.load(filename), object))
        os.remove(filename)
        
    
    def test_tools_noise_brown(self):
        ncols=3
        nrows=2
        self.assertEqual(LFPy.tools.noise_brown(ncols, nrows).shape, (nrows, ncols))


#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
An LFPy example file showing how cells can be run in parallel using MPI.
To run using MPI with 4 cpu cores, issue in terminal
openmpirun -np 4 python example_mpi.py

The example uses mpi4py with openmpi, and do not rely on NEURON's MPI
implementation.

Execution:

    <mpiexec> -n <processes> python example_mpi.py

Notes:
- on certain platforms and with mpirun, the --oversubscribe argument is needed
  to get more processes than the number of physical CPU cores.

Copyright (C) 2017 Computational Neuroscience Group, NMBU.

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
"""

import os
from os.path import join
import numpy as np
import scipy.stats as st
import matplotlib.pyplot as plt
from matplotlib.collections import PolyCollection
import LFPy
import neuron
import sys
if sys.version < '3':
    from urllib2 import urlopen
else:    
    from urllib.request import urlopen
import zipfile
from warnings import warn
from mpi4py import MPI

#MPI stuff we're using
COMM = MPI.COMM_WORLD
SIZE = COMM.Get_size()
RANK = COMM.Get_rank()

#Fetch Mainen&Sejnowski 1996 model files
if not os.path.isfile(join('cells', 'cells', 'j4a.hoc')) and RANK == 0:

    #get the model files:
    u = urlopen('http://senselab.med.yale.edu/ModelDB/eavBinDown.asp?o=2488&a=23&mime=application/zip')
    localFile = open('patdemo.zip', 'w')
    localFile.write(u.read())
    localFile.close()
    #unzip:
    myzip = zipfile.ZipFile('patdemo.zip', 'r')
    myzip.extractall('.')
    myzip.close()

#compile mod files every time, because of incompatibility with Hay2011 files:
if "win32" in sys.platform:
    pth = "cells"
    warn("no autompile of NMODL (.mod) files on Windows. " 
         + "Run mknrndll from NEURON bash in the folder cells and rerun example script")
    if not pth in neuron.nrn_dll_loaded:
        neuron.h.nrn_load_dll(pth+"/nrnmech.dll")
    neuron.nrn_dll_loaded.append(pth)
else:
    if RANK == 0:
        os.system('''
                  cd cells
                  nrnivmodl
                  ''')
    COMM.Barrier()
    neuron.load_mechanisms('cells')


    

#set one global seed, ensure all randomizations are set on RANK 0 in script!
np.random.seed(12345)


class Population:
    '''prototype population class'''
    def __init__(self, POPULATION_SIZE,
                 cellParameters,
                 populationParameters,
                 electrodeParameters,
                 synapseParameters,
                 stationaryGammaArgs,
                 ):
        '''
        class initialization
        
        POPULATION_SIZE :       int, number of cells
        cellParameters :        dict
        populationParameters :  dict
        electrodeParameters :   dict
        synapseParameters :     dict
        stationaryGammaArgs :   dict
        
        '''
        self.POPULATION_SIZE = POPULATION_SIZE
        self.cellParameters = cellParameters
        self.populationParameters = populationParameters
        self.electrodeParameters = electrodeParameters
        self.synapseParameters = synapseParameters
        self.stationaryGammaArgs = stationaryGammaArgs
        
        #get synaptic times and cell positions, rotations, store in self-object
        self.synapseTimes = self.drawRandSpikeTimes()
        self.cellPositions = self.drawRandCellPositions()
        self.cellRotations = self.drawRandCellRotations()
        
    def run(self):
        '''execute the proper simulation and collect simulation results'''
        #produce simulation results on each RANK
        self.results = self.distribute_cellsims()
        
        #superimpose local LFPs on every RANK, then sum using MPI to RANK 0
        self.LFP = []
        for key, value in list(self.results.items()):
            self.LFP.append(value['LFP'])
        self.LFP = np.array(self.LFP).sum(axis=0)
        self.LFP = COMM.reduce(self.LFP)        #LFP is None on all but RANK 0
        
        #collect all simulation results on RANK 0, including single cell LFP
        if RANK == 0:
            for i in range(1, SIZE):
                result = COMM.recv(source=MPI.ANY_SOURCE) #receive from ANY rank
                self.results.update(result)     #collect
        else:
            COMM.send(self.results, dest=0)     #send to RANK 0
            self.results = None                 #results only exist on RANK 0
            
        COMM.Barrier()  #sync MPI threads
    
    def distribute_cellsims(self):
        '''Will distribute and run cell simulations across ranks'''
        #start unique cell simulation on every RANK,
        #and store the electrode and cell objects in dicts indexed by cellindex
        results = {}
        for cellindex in range(self.POPULATION_SIZE):
            if divmod(cellindex, SIZE)[1] == RANK:
                results.update({cellindex : self.cellsim(cellindex)})
        return results

    def drawRandSpikeTimes(self):
        '''draw and distribute some spike times for each cell in population'''
        if RANK == 0:
            RandSpikeTimes = []
            for cellindex in range(self.POPULATION_SIZE):
                sptimes = LFPy.inputgenerators.get_activation_times_from_distribution(
                    n=1, tstart=0, tstop=cellParameters['tstop'],
                    distribution=st.gamma,
                    rvs_args = self.stationaryGammaArgs
                )
                RandSpikeTimes.append(sptimes[0])
        else:
            RandSpikeTimes = None
        return COMM.bcast(RandSpikeTimes, root=0)

    def drawRandCellPositions(self):
        '''draw and distribute some random cell positions within a
        cylinder constraints'''
        if RANK == 0:
            cellPositions = []
            for cellindex in range(self.POPULATION_SIZE):
                r = np.sqrt(np.random.rand()) * \
                                    self.populationParameters['radius']
                theta = np.random.rand() * 2 * np.pi
                x = r * np.sin(theta)
                y = r * np.cos(theta)
                z = np.random.rand() * (self.populationParameters['zmax'] -
                                        self.populationParameters['zmin']) + \
                                    self.populationParameters['zmin']
                cellPositions.append([x, y, z])
            cellPositions = np.array(cellPositions)
        else:
            cellPositions = None
        return COMM.bcast(cellPositions, root=0)
    
    def drawRandCellRotations(self):
        '''draw and distribute random cell rotations for all cells in population
        '''
        if RANK == 0:
            cellRotations = np.random.rand(self.POPULATION_SIZE) * np.pi * 2
        else:
            cellRotations = None
        return COMM.bcast(cellRotations, root=0)
        
    def cellsim(self, cellindex):
        '''main cell- and LFP simulation procedure'''
        #create extracellular electrode object
        electrode = LFPy.RecExtElectrode(**self.electrodeParameters)
        
        #Initialize cell instance, using the LFPy.Cell class
        cell = LFPy.Cell(**self.cellParameters)
        #set the position of midpoint in soma
        cell.set_pos(x = self.cellPositions[cellindex, 0],
                     y = self.cellPositions[cellindex, 1],
                     z = self.cellPositions[cellindex, 2])
        #rotate the morphology
        cell.set_rotation(z = self.cellRotations[cellindex])
        
        #attach synapse with parameters and set spike time
        synapse = LFPy.Synapse(cell, **self.synapseParameters)
        synapse.set_spike_times(self.synapseTimes[cellindex])
        
        #perform NEURON simulation, results saved as attributes in cell
        cell.simulate(electrode = electrode)
        
        #return dict with primary results from simulation
        return {'LFP' : electrode.LFP, 'somav' : cell.somav}

    def plotstuff(self):
        '''plot LFPs and somatraces'''
                
        if RANK == 0:
            fig = plt.figure(figsize=(12, 8))
            
            ax = fig.add_axes([0.05, 0.0, 0.45, 1.0],
                        aspect='equal', frameon=False,
                        xticks=[], xticklabels=[], yticks=[], yticklabels=[])
            for cellindex in range(self.POPULATION_SIZE):
                cell = LFPy.Cell(**self.cellParameters)
                cell.set_pos(x = self.cellPositions[cellindex, 0],
                             y = self.cellPositions[cellindex, 1],
                             z = self.cellPositions[cellindex, 2])
                cell.set_rotation(z = self.cellRotations[cellindex])

                zips = []
                for x, z in cell.get_idx_polygons():
                    zips.append(list(zip(x, z)))
                
                polycol = PolyCollection(zips,
                                edgecolors='none',
                                facecolors='bgrcmykbgrcmykbgrcmyk'[cellindex],
                                zorder = self.cellPositions[cellindex, 1])
       
                ax.add_collection(polycol)

            ax.plot(self.electrodeParameters['x'],
                    self.electrodeParameters['z'],
                    marker='o', color='g', clip_on=False, zorder=0)
            
            ax = fig.add_axes([0.5, 0.55, 0.40, 0.4])
            for key, value in list(self.results.items()):
                tvec = np.arange(value['somav'].size) * \
                                        self.cellParameters['dt']
                ax.plot(tvec, value['somav'],
                        label = 'cell %i' % key)
            leg = ax.legend()
            #ax.set_xlabel('time (ms)')
            ax.set_ylabel('$V_{soma}$ (mV)')
            ax.set_title('somatic potentials')
            
            ax = fig.add_axes([0.5, 0.075, 0.40, 0.4])
            cax = fig.add_axes([0.91, 0.075, 0.02, 0.40])
            im = ax.pcolormesh(tvec, self.electrodeParameters['z'], self.LFP,
                           cmap='PRGn',
                           vmin = -self.LFP.std()*3,
                           vmax = self.LFP.std()*3)
            ax.axis(ax.axis('tight'))
            cbar = plt.colorbar(im, cax=cax)
            cbar.set_label('LFP (mV)')
            ax.set_title('superimposed LFP')
            ax.set_xlabel('time (ms)')
            ax.set_ylabel('$z$ ($\mu$m)')
            
            fig.savefig('example_mpi.pdf', dpi=300)
            plt.show()
        

if __name__ == '__main__':
    ########### PARAMETERS ##################
    #define cell parameters used as input to cell-class
    cellParameters = {
        'morphology' : join('morphologies', 'L5_Mainen96_wAxon_LFPy.hoc'),
        'cm' : 1.0,                 # membrane capacitance
        'Ra' : 150,                 # axial resistance
        'v_init' : -65,             # initial crossmembrane potential
        'passive' : True,           # switch on passive mechs
        'passive_parameters' : {'g_pas' : 1./30000, 'e_pas' : -65}, # passive params
        'nsegs_method' : 'lambda_f',# method for setting number of segments,
        'lambda_f' : 10,            # segments are isopotential at frequency
        'dt' : 2**-3,               # dt of LFP and NEURON simulation.
        'tstart' : -100,          #start time, recorders start at t=0
        'tstop' : 200,            #stop time of simulation
        'custom_code'  : ['active_declarations_example3.hoc'], #active decl.
    }
    
    #Synaptic parameters, corresponding to a NetCon synapse built into NEURON
    synapseParameters = {
        'idx' : 0,               # insert synapse on index "0", the soma
        'e' : 0.,                # reversal potential of synapse
        'syntype' : 'Exp2Syn',   # conductance based double-exponential synapse
        'tau1' : 1.0,            # Time constant, rise
        'tau2' : 1.0,            # Time constant, decay
        'weight' : 0.002,        # Synaptic weight
        'record_current' : False,# disable synapse current recording
    }
    
    #parameter args for LFPy.inputgenerators.stationary_gamma()
    stationaryGammaArgs = {
        'a' : 0.25,               #shape parameter
        'scale' : 12,             #"rate" parameter
    }
    
    # Define electrode geometry corresponding to a laminar electrode, where
    # contact points have a radius r, surface normal vectors N, and LFP
    # calculated as the average LFP in n random points on each contact:
    X, Y, Z = np.mgrid[0:1, 0:1, -500:1001:50]
    N = np.zeros((X.size, 3))
    for i in range(N.shape[0]): N[i,] = [1, 0, 0] #normal unit vec. to contacts
    # put parameters in dictionary
    electrodeParameters = {
        'sigma' : 0.3,              # Extracellular potential
        'x' : X.flatten(),      # x,y,z-coordinates of electrode contacts
        'y' : Y.flatten(),
        'z' : Z.flatten(),
        'n' : 50,
        'r' : 5,
        'N' : N,
    }
    
    #the number of cells in the population
    POPULATION_SIZE = 4
    
    #will draw random cell locations within cylinder constraints:
    populationParameters = {
        'radius' : 20,
        'zmin' : -200,
        'zmax' : 200,
    }
    
    ########## INITIALIZE POPULATION ###########################################
    population = Population(POPULATION_SIZE,
                     cellParameters,
                     populationParameters,
                     electrodeParameters,
                     synapseParameters,
                     stationaryGammaArgs,)
    population.run()
    population.plotstuff()
    



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Example plot for LFPy: Single-synapse contribution to the LFP

Copyright (C) 2017 Computational Neuroscience Group, NMBU.

Execution:

    python example_anisotropy.py
    
This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
"""
import LFPy
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.collections import PolyCollection
from os.path import join


cell_parameters = {          # various cell parameters,
    'morphology' : join('morphologies', 'ball_and_stick.hoc'), # Mainen&Sejnowski, 1996
    'passive_parameters': {"g_pas": 1/30000., "e_pas": -70},
    'cm' : 1.0,         # membrane capacitance
    'Ra' : 150,        # axial resistance
    'v_init' : -65.,    # initial crossmembrane potential
    'passive' : True,   # switch on passive mechs
    'nsegs_method' : 'lambda_f',
    'lambda_f' : 1000.,
    'dt' : 2.**-2,      # simulation time step size
    'tstart' : -10.,    # start time of simulation, recorders start at t=0
    'tstop' : 10.,   # stop simulation at 200 ms. These can be overridden
                        # by setting these arguments i cell.simulation()
}

# Create cell
cell = LFPy.Cell(**cell_parameters)
cell.set_pos(z=-10)

synapse_parameters = {
    'idx' : 0,
    'e' : 0.,                   # reversal potential
    'syntype' : 'ExpSyn',       # synapse type
    'tau' : 5.,                 # synaptic time constant
    'weight' : .001,            # synaptic weight
    'record_current' : False,    # record synapse current
}

# Create synapse and set time of synaptic input
synapse = LFPy.Synapse(cell, **synapse_parameters)
synapse.set_spike_times(np.array([5.]))


# Create a grid of measurement locations, in (um)
X, Z = np.mgrid[-100:101:2, -100:200:2]
Y = np.zeros(X.shape)


sigma = 0.3
sigma_tensor = [0.3, 0.3, 0.45]

# Define electrode parameters
grid_electrode_parameters = {
    'sigma' : sigma,      # extracellular conductivity
    'x' : X.flatten(),  # electrode requires 1d vector of positions
    'y' : Y.flatten(),
    'z' : Z.flatten(),
    'method': 'soma_as_point'
}

grid_electrode_parameters_tensor = {
    'sigma' : sigma_tensor,      # extracellular conductivity
    'x' : X.flatten(),  # electrode requires 1d vector of positions
    'y' : Y.flatten(),
    'z' : Z.flatten(),
    'method': 'soma_as_point'
}


# Run simulation, electrode object argument in cell.simulate
print("running simulation...")

cell.simulate(rec_imem=True)

# Create electrode objects

grid_electrode = LFPy.RecExtElectrode(cell, **grid_electrode_parameters)
grid_electrode.calc_lfp()

grid_electrode_tensor = LFPy.RecExtElectrode(cell, **grid_electrode_parameters_tensor)
grid_electrode_tensor.calc_lfp()

fig = plt.figure(figsize=[10, 5])

ax = fig.add_subplot(121, aspect='equal', xlabel='x ($\mu$m)', ylabel='z ($\mu$m)', title="Sigma: %s S/m" % str(sigma),
                     ylim=[np.min(grid_electrode.z), np.max(grid_electrode.z)],
                     xlim=[np.min(grid_electrode.x), np.max(grid_electrode.x)])

max_idx = np.argmax(np.abs(cell.imem[0,:]))

LFP = 1000 * grid_electrode.LFP[:,max_idx].reshape(X.shape)
im = ax.contourf(X, Z, LFP, 51, vmin=-np.max(np.abs(LFP)) / 1, vmax=np.max(np.abs(LFP)) / 1,
           cmap='bwr',
           zorder=-2)
ax.contour(X, Z, LFP, 51, vmin=-np.max(np.abs(LFP)) / 1, vmax=np.max(np.abs(LFP)) / 1,
           cmap='bwr',
           zorder=-2)

cbar = plt.colorbar(im)
cbar.set_label('LFP ($\mu$V)')

#plot morphology
zips = []
for x, z in cell.get_idx_polygons():
    zips.append(list(zip(x, z)))
polycol = PolyCollection(zips, alpha=0.2,
                         edgecolors='none',
                         facecolors='k')
ax.add_collection(polycol)


ax.plot(cell.xmid[cell.synidx],cell.zmid[cell.synidx], 'o', ms=5,
        markeredgecolor='k',
        markerfacecolor='r')

ax2 = fig.add_subplot(122, aspect='equal', xlabel='x ($\mu$m)', title="Sigma: %s S/m" % str(sigma_tensor),
                     ylim=[np.min(grid_electrode.z), np.max(grid_electrode.z)],
                     xlim=[np.min(grid_electrode.x), np.max(grid_electrode.x)])

LFP = 1000 * grid_electrode_tensor.LFP[:,max_idx].reshape(X.shape)
im = ax2.contourf(X, Z, LFP, 51, vmin=-np.max(np.abs(LFP)) / 1, vmax=np.max(np.abs(LFP)) / 1,
           cmap='bwr',
           zorder=-2)
ax2.contour(X, Z, LFP, 51, vmin=-np.max(np.abs(LFP)) / 1, vmax=np.max(np.abs(LFP)) / 1,
           cmap='bwr',
           zorder=-2)

cbar = plt.colorbar(im)
cbar.set_label('LFP ($\mu$V)')
#plot morphology
zips = []
for x, z in cell.get_idx_polygons():
    zips.append(list(zip(x, z)))
polycol = PolyCollection(zips,
                         edgecolors='none', alpha=0.2,
                         facecolors='k')
ax2.add_collection(polycol)


ax2.plot(cell.xmid[cell.synidx],cell.zmid[cell.synidx], 'o', ms=5,
        markeredgecolor='k',
        markerfacecolor='r')

plt.savefig('example_anisotropy.pdf', dpi=150)

plt.show()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Copyright (C) 2012 Computational Neuroscience Group, NMBU.

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

"""


import os
import unittest
import numpy as np
import LFPy
import neuron
import h5py
import scipy.signal as ss

class testNetworkPopulation(unittest.TestCase):
    """
    class LFPy.NetworkPopulation test suite
    """
    def test_NetworkPopulation_00(self):
        cellParameters = dict(
            morphology=os.path.join(LFPy.__path__[0], 'test', 'ball_and_sticks_w_lists.hoc'),
            templatefile=os.path.join(LFPy.__path__[0], 'test', 'ball_and_stick_template.hoc'),
            templatename='ball_and_stick_template',
            templateargs=None,
            passive=False,
            dt=2**-3,
            tstop=100,
            delete_sections=False,
        )

        populationParameters = dict(
            CWD=None,
            CELLPATH=None,
            Cell=LFPy.NetworkCell,
            cell_args = cellParameters,
            pop_args = dict(
                radius=100,
                loc=0.,
                scale=20.),
            rotation_args = dict(x=0, y=0),
            POP_SIZE = 4,
            name = 'ball_and_sticks',
            OUTPUTPATH='tmp_testNetworkPopulation'
        )

        population = LFPy.NetworkPopulation(**populationParameters)

        self.assertTrue(len(population.cells) == population.POP_SIZE)
        for cell, soma_pos, gid in zip(population.cells, population.soma_pos, population.gids):
            self.assertTrue(type(cell) is LFPy.NetworkCell)
            self.assertTrue((cell.somapos[0] == soma_pos['x']) &
                            (cell.somapos[1] == soma_pos['y']) &
                            (cell.somapos[2] == soma_pos['z']))
            self.assertEqual(cell.gid, gid)
            self.assertTrue(np.sqrt(soma_pos['x']**2 + soma_pos['y']**2) <= 100.)
        np.testing.assert_equal(population.gids, np.arange(4))


        os.system('rm -r tmp_testNetworkPopulation')
        neuron.h('forall delete_section()')


class testNetwork(unittest.TestCase):
    """
    class LFPy.Network test suite
    """
    def test_Network_00(self):
        cellParameters = dict(
            morphology=os.path.join(LFPy.__path__[0], 'test', 'ball_and_sticks_w_lists.hoc'),
            templatefile=os.path.join(LFPy.__path__[0], 'test', 'ball_and_stick_template.hoc'),
            templatename='ball_and_stick_template',
            templateargs=None,
            passive=False,
            dt=2**-3,
            tstop=100,
            delete_sections=False,
        )

        populationParameters = dict(
            CWD=None,
            CELLPATH=None,
            Cell=LFPy.NetworkCell,
            cell_args = cellParameters,
            pop_args = dict(
                radius=100,
                loc=0.,
                scale=20.),
            rotation_args = dict(x=0, y=0),
            POP_SIZE = 4,
            name = 'test',
        )
        networkParameters = dict(
            dt=2**-3,
            tstart=0.,
            tstop=100.,
            v_init=-65.,
            celsius=6.3,
            OUTPUTPATH='tmp_testNetworkPopulation'
            )
        # set up
        network = LFPy.Network(**networkParameters)
        network.create_population(**populationParameters)
        connectivity = network.get_connectivity_rand(pre='test', post='test', connprob=0.5)

        # test set up
        for population in network.populations.values():
            self.assertTrue(len(population.cells) == population.POP_SIZE)
            for cell, soma_pos, gid in zip(population.cells, population.soma_pos, population.gids):
                self.assertTrue(type(cell) is LFPy.NetworkCell)
                self.assertTrue((cell.somapos[0] == soma_pos['x']) &
                                (cell.somapos[1] == soma_pos['y']) &
                                (cell.somapos[2] == soma_pos['z']))
                self.assertEqual(cell.gid, gid)
                self.assertTrue(np.sqrt(soma_pos['x']**2 + soma_pos['y']**2) <= 100.)
            np.testing.assert_equal(population.gids, np.arange(4))

        np.testing.assert_equal(connectivity.shape, (population.POP_SIZE, population.POP_SIZE))
        np.testing.assert_equal(connectivity.diagonal(), np.zeros(population.POP_SIZE))

        # connect and run sim
        network.connect(pre='test', post='test', connectivity=connectivity)
        network.simulate()

        # test output
        for population in network.populations.values():
            for cell in population.cells:
                self.assertTrue(np.all(cell.somav == network.v_init))

        network.pc.gid_clear()
        os.system('rm -r tmp_testNetworkPopulation')
        neuron.h('forall delete_section()')


    def test_Network_01(self):
        cellParameters = dict(
            morphology=os.path.join(LFPy.__path__[0], 'test', 'ball_and_sticks_w_lists.hoc'),
            templatefile=os.path.join(LFPy.__path__[0], 'test', 'ball_and_stick_template.hoc'),
            templatename='ball_and_stick_template',
            templateargs=None,
            passive=False,
            dt=2**-3,
            tstop=100,
            delete_sections=False,
        )

        populationParameters = dict(
            CWD=None,
            CELLPATH=None,
            Cell=LFPy.NetworkCell,
            cell_args = cellParameters,
            pop_args = dict(
                radius=100,
                loc=0.,
                scale=20.),
            rotation_args = dict(x=0, y=0),
            POP_SIZE = 4,
            name = 'test',
        )
        networkParameters = dict(
            dt=2**-3,
            tstart=0.,
            tstop=100.,
            v_init=-65.,
            celsius=6.3,
            OUTPUTPATH='tmp_testNetworkPopulation'
            )
        # set up
        network = LFPy.Network(**networkParameters)
        network.create_population(**populationParameters)
        connectivity = network.get_connectivity_rand(pre='test', post='test', connprob=0.5)

        # connect and run sim
        network.connect(pre='test', post='test', connectivity=connectivity)
        SPIKES, LFP, P = network.simulate(rec_current_dipole_moment=True)

        # test output
        for population in network.populations.values():
            for cell in population.cells:
                self.assertTrue(np.all(cell.somav == network.v_init))

        self.assertTrue(np.all(P['test'] == 0.))
        self.assertTrue(P.shape[0] == cell.somav.size)
        self.assertTrue(len(LFP) == 0)

        network.pc.gid_clear()
        os.system('rm -r tmp_testNetworkPopulation')
        neuron.h('forall delete_section()')


    def test_Network_02(self):
        cellParameters = dict(
            morphology=os.path.join(LFPy.__path__[0], 'test', 'ball_and_sticks_w_lists.hoc'),
            templatefile=os.path.join(LFPy.__path__[0], 'test', 'ball_and_stick_template.hoc'),
            templatename='ball_and_stick_template',
            templateargs=None,
            passive=False,
            dt=2**-3,
            tstop=100,
            delete_sections=False,
        )

        populationParameters = dict(
            CWD=None,
            CELLPATH=None,
            Cell=LFPy.NetworkCell,
            cell_args = cellParameters,
            pop_args = dict(
                radius=100,
                loc=0.,
                scale=20.),
            rotation_args = dict(x=0, y=0),
            POP_SIZE = 4,
            name = 'test',
        )
        networkParameters = dict(
            dt=2**-3,
            tstart=0.,
            tstop=100.,
            v_init=-65.,
            celsius=6.3,
            OUTPUTPATH='tmp_testNetworkPopulation'
            )
        clampParams = {
            'idx' : 0,
            'pptype' : 'VClamp',
            'amp[0]' : -65,
            'dur[0]' : 10,
            'amp[1]' : 0,
            'dur[1]' : 1,
            'amp[2]' : -65,
            'dur[2]' : 1E8,
        }

        # set up
        network = LFPy.Network(**networkParameters)
        network.create_population(**populationParameters)
        connectivity = network.get_connectivity_rand(pre='test', post='test', connprob=1)

        # test connectivity
        self.assertTrue(np.all(connectivity == (np.eye(populationParameters['POP_SIZE']) == 0)))

        # connect
        network.connect(pre='test', post='test', connectivity=connectivity,
                        multapseargs=dict(loc=1, scale=1E-9))

        # create synthetic AP in cell with gid == 0
        for population in network.populations.values():
            for cell in population.cells:
                if cell.gid == 0:
                    vclamp = LFPy.StimIntElectrode(cell=cell, **clampParams)

        # simulate
        network.simulate()

        # test output
        for population in network.populations.values():
            for cell in population.cells:
                self.assertFalse(np.all(cell.somav == network.v_init))

        network.pc.gid_clear()
        os.system('rm -r tmp_testNetworkPopulation')
        neuron.h('forall delete_section()')


    def test_Network_03(self):
        cellParameters = dict(
            morphology=os.path.join(LFPy.__path__[0], 'test', 'ball_and_sticks_w_lists.hoc'),
            templatefile=os.path.join(LFPy.__path__[0], 'test', 'ball_and_stick_template.hoc'),
            templatename='ball_and_stick_template',
            templateargs=None,
            passive=False,
            dt=2**-3,
            tstop=100,
            delete_sections=False,
        )

        populationParameters = dict(
            CWD=None,
            CELLPATH=None,
            Cell=LFPy.NetworkCell,
            cell_args = cellParameters,
            pop_args = dict(
                radius=100,
                loc=0.,
                scale=20.),
            rotation_args = dict(x=0, y=0),
            POP_SIZE = 4,
            name = 'test',
        )
        networkParameters = dict(
            dt=2**-3,
            tstart=0.,
            tstop=100.,
            v_init=-65.,
            celsius=6.3,
            OUTPUTPATH='tmp_testNetworkPopulation'
            )
        electrodeParameters = dict(
            sigma=0.3,
            x = np.arange(10)*100,
            y = np.arange(10)*100,
            z = np.arange(10)*100
        )
        # set up
        network = LFPy.Network(**networkParameters)
        network.create_population(**populationParameters)
        connectivity = network.get_connectivity_rand(pre='test', post='test', connprob=0.5)

        # test set up
        for population in network.populations.values():
            self.assertTrue(len(population.cells) == population.POP_SIZE)
            for cell, soma_pos, gid in zip(population.cells, population.soma_pos, population.gids):
                self.assertTrue(type(cell) is LFPy.NetworkCell)
                self.assertTrue((cell.somapos[0] == soma_pos['x']) &
                                (cell.somapos[1] == soma_pos['y']) &
                                (cell.somapos[2] == soma_pos['z']))
                self.assertEqual(cell.gid, gid)
                self.assertTrue(np.sqrt(soma_pos['x']**2 + soma_pos['y']**2) <= 100.)
            np.testing.assert_equal(population.gids, np.arange(4))

        np.testing.assert_equal(connectivity.shape, (population.POP_SIZE, population.POP_SIZE))
        np.testing.assert_equal(connectivity.diagonal(), np.zeros(population.POP_SIZE))

        # set up electrode
        electrode = LFPy.RecExtElectrode(**electrodeParameters)

        # connect and run sim
        network.connect(pre='test', post='test', connectivity=connectivity)
        network.simulate(electrode=electrode, to_file=True, to_memory=False,
                         file_name='OUTPUT.h5')

        # test output
        for population in network.populations.values():
            for cell in population.cells:
                self.assertTrue(np.all(cell.somav == network.v_init))
            
        f = h5py.File(os.path.join(network.OUTPUTPATH, 'OUTPUT.h5'), 'r')        
        np.testing.assert_equal(f['OUTPUT[0]'][()], np.zeros_like(f['OUTPUT[0]'][()]))
        f.close()


        network.pc.gid_clear()
        os.system('rm -r tmp_testNetworkPopulation')
        neuron.h('forall delete_section()')


    def test_Network_04(self):
        cellParameters = dict(
            morphology=os.path.join(LFPy.__path__[0], 'test', 'ball_and_sticks_w_lists.hoc'),
            templatefile=os.path.join(LFPy.__path__[0], 'test', 'ball_and_stick_template.hoc'),
            templatename='ball_and_stick_template',
            templateargs=None,
            passive=True,
            dt=2**-3,
            tstop=100,
            delete_sections=False,
        )
        
        synapseParameters = dict(idx=0, syntype='Exp2Syn', weight=0.002,
                                 tau1=0.1, tau2=0.1, e=0)
        
        populationParameters = dict(
            CWD=None,
            CELLPATH=None,
            Cell=LFPy.NetworkCell,
            cell_args = cellParameters,
            pop_args = dict(
                radius=100,
                loc=0.,
                scale=20.),
            rotation_args = dict(x=0, y=0),
            POP_SIZE = 1,
            name = 'test',
        )
        networkParameters = dict(
            dt=2**-3,
            tstart=0.,
            tstop=100.,
            v_init=-70.,
            celsius=6.3,
            OUTPUTPATH='tmp_testNetworkPopulation'
            )
        # set up
        network = LFPy.Network(**networkParameters)
        network.create_population(**populationParameters)
        
        cell = network.populations['test'].cells[0]
        
        # create synapses
        synlist = []
        numsynapses = 2
        for i in range(numsynapses):
            synlist.append(LFPy.Synapse(cell=cell, **synapseParameters))
            synlist[-1].set_spike_times(np.array([10+(i*10)]))
        
        network.simulate()
        
        # test that the input results in the correct amount of PSPs
        np.testing.assert_equal(ss.argrelextrema(cell.somav, np.greater)[0].size, numsynapses)

        # clean up
        network.pc.gid_clear()
        os.system('rm -r tmp_testNetworkPopulation')
        neuron.h('forall delete_section()')


    def test_Network_05(self):
        cellParameters = dict(
            morphology=os.path.join(LFPy.__path__[0], 'test', 'ball_and_sticks_w_lists.hoc'),
            templatefile=os.path.join(LFPy.__path__[0], 'test', 'ball_and_stick_template.hoc'),
            templatename='ball_and_stick_template',
            templateargs=None,
            passive=False,
            dt=2**-3,
            tstop=100,
            delete_sections=False,
        )

        populationParameters = dict(
            CWD=None,
            CELLPATH=None,
            Cell=LFPy.NetworkCell,
            cell_args = cellParameters,
            pop_args = dict(
                radius=100,
                loc=0.,
                scale=20.),
            rotation_args = dict(x=0, y=0),
            POP_SIZE = 4,
            name = 'test',
        )
        networkParameters = dict(
            dt=2**-3,
            tstart=0.,
            tstop=100.,
            v_init=-65.,
            celsius=6.3,
            OUTPUTPATH='tmp_testNetworkPopulation'
            )
        electrodeParameters = dict(
            sigma=0.3,
            x = np.arange(10)*100,
            y = np.arange(10)*100,
            z = np.arange(10)*100
        )
        # set up
        network = LFPy.Network(**networkParameters)
        network.create_population(**populationParameters)
        connectivity = network.get_connectivity_rand(pre='test', post='test',
                                                     connprob=0.5)

        # test set up
        for population in network.populations.values():
            self.assertTrue(len(population.cells) == population.POP_SIZE)
            for cell, soma_pos, gid in zip(population.cells,
                                           population.soma_pos,
                                           population.gids):
                self.assertTrue(type(cell) is LFPy.NetworkCell)
                self.assertTrue((cell.somapos[0] == soma_pos['x']) &
                                (cell.somapos[1] == soma_pos['y']) &
                                (cell.somapos[2] == soma_pos['z']))
                self.assertEqual(cell.gid, gid)
                self.assertTrue(np.sqrt(soma_pos['x']**2 + soma_pos['y']**2) <= 100.)
            np.testing.assert_equal(population.gids, np.arange(4))

        np.testing.assert_equal(connectivity.shape,
                                (population.POP_SIZE, population.POP_SIZE))
        np.testing.assert_equal(connectivity.diagonal(),
                                np.zeros(population.POP_SIZE))

        # set up electrode
        electrode = LFPy.RecExtElectrode(**electrodeParameters)

        # connect and run sim
        network.connect(pre='test', post='test', connectivity=connectivity)
        SPIKES, RESULTS, P = network.simulate(electrode=electrode, to_file=True,
                                      to_memory=True, file_name='OUTPUT.h5')

        # test output
        for population in network.populations.values():
            for cell in population.cells:
                self.assertTrue(np.all(cell.somav == network.v_init))
            
        f = h5py.File(os.path.join(network.OUTPUTPATH, 'OUTPUT.h5'), 'r')        
        np.testing.assert_equal(f['OUTPUT[0]'][()], RESULTS[0])
        f.close()


        network.pc.gid_clear()
        os.system('rm -r tmp_testNetworkPopulation')
        neuron.h('forall delete_section()')

